#
"""
tests for code in lib/utils/jsonutils to convert Python/Larch
objects to and from representations that can be serialized with JSON.
"""
from __future__ import print_function
import unittest
import numpy as np
from numpy.testing import assert_allclose

from larch import isParameter, Parameter, isgroup, Group
from larch.utils.jsonutils import encode4js, decode4js

def encode(obj):
    out = encode4js(obj)
    print("Encoded to: ", out)
    return out

def deval(obj):
    out = decode4js(encode(obj))
    print("Decoded to: ", out)
    return out

def evaltest(obj):
    assert(deval(obj) == obj)

class EncodeDecode4Json_Test(unittest.TestCase):
    def setUp(self):
        self.param1 = Parameter(name='a', value=2.0, vary=True, min=0)
        self.param2 = Parameter(name='b', expr='sqrt(x/2)')
        self.group1 = Group(a     = 1.030405,
                            label = 'a string',
                            opts  = {'a':1, 'b':2},
                            x     = np.linspace(0, 10, 21),
                            y     = np.sqrt(np.arange(3)/5.))

        self.group2 = Group(par1 = Parameter(name='p1', value=3.0, min=0.0, vary=False),
                            par2 = Parameter(name='p2', value=1.0, vary=True),
                            sub  = Group(label='a label',
                                         x = np.linspace(0, 10, 21)))

        self.group3 = Group(par1 = Parameter(name='p1', value=3.0, min=0.0),
                            dx    = {'a': 1.3+0.2j, 'blist': [1,2,3.0],
                                     'tup': (0, None)} )


    def test_encode_string(self):
        out = encode('hello')
        assert(out == 'hello')

    def test_eval_string(self):
        evaltest('hello')

    def test_encode_int(self):
        out = encode(99)
        assert(out == 99)

    def test_eval_int(self):
        evaltest(99)

    def test_encode_double(self):
        out = encode(2.0)
        assert(out == 2.0)

    def test_eval_double(self):
        evaltest(9.95)


    def test_encode_complex(self):
        out = encode(2.0 + 1.0j)
        assert(isinstance(out, dict))
        assert(out['__class__'] == 'Complex')
        assert(out['value'] == (2.0, 1.0))

    def test_eval_complex(self):
        evaltest(2.103 + 99.11j)

    def test_encode_list(self):
        out = encode([1, 2, 'a', ['x', 'y', 'z']])
        assert(isinstance(out, dict))
        assert(out['__class__'] == 'List')
        assert(out['value'][0] == 1)
        assert(out['value'][2] == 'a')
        assert(isinstance(out['value'][3], dict))
        assert(out['value'][3]['__class__'] == 'List')
        assert(out['value'][3]['value'][1] == 'y')

    def test_eval_list(self):
        evaltest( [1, 2, 'a', ['x', 'y', 'z']])

    def test_encode_tuple(self):
        out = encode(('s', None, False))
        assert(isinstance(out, dict))
        assert(out['__class__'] == 'Tuple')
        assert(out['value'][0] == 's')
        assert(out['value'][1] == None)
        assert(out['value'][2] == False)

    def test_eval_tuple(self):
        evaltest(('s', None, False, ('a', 'b', 'c')))

    def test_encode_array1(self):
        out = encode(np.arange(10)/7.0)
        assert(isinstance(out, dict))
        assert(out['__class__'] == 'Array')
        assert(out['__shape__'] == (10, ))
        assert(out['value'][0] == 0)
        assert(len(out['value']) == 10)
        assert_allclose(out['value'][1], 0.142857, rtol=1.e-4)

    def test_eval_array1(self):
        testval = np.arange(10)/7.2
        assert_allclose(deval(testval), testval, rtol=1.e-4)

    def test_encode_array2(self):
        out = encode((np.arange(20)/2.0).reshape((4, 5)))
        assert(isinstance(out, dict))
        assert(out['__class__'] == 'Array')
        assert(out['__shape__'] == (4, 5))
        assert(len(out['value']) == 20)
        arr = np.array(out['value']).reshape(out['__shape__'])
        assert_allclose(arr[1][2], 3.5, rtol=1.e-4)

    def test_eval_array2(self):
        testval = (np.arange(20)/7.2).reshape((4, 5))
        out = deval(testval)
        assert(out.shape == (4, 5))
        assert_allclose(out, testval, rtol=1.e-4)

    def test_encode_array3(self):
        out = encode(np.arange(10)/2.0 - 1j*np.arange(10))
        assert(isinstance(out, dict))
        assert(out['__class__'] == 'Array')
        assert(out['__shape__'] == (10,))
        assert(out['__dtype__'].startswith('complex'))
        assert(len(out['value']) == 2)
        val = out['value']
        assert(isinstance(val, list))
        assert(len(val)    == 2)
        assert(len(val[0]) == 10)
        assert_allclose(val[0][4],  2.0, rtol=1.e-4)
        assert_allclose(val[1][3], -3.0, rtol=1.e-4)

    def test_eval_array3(self):
        testval = np.arange(10)/2.0 - 1j*np.arange(10)
        out = deval(testval)
        assert(out.dtype == complex)
        assert_allclose(out, testval, rtol=1.e-4)

    def test_encode_param1(self):
        out = encode(self.param1)
        assert(isinstance(out, dict))
        assert(out['__class__'] == 'Parameter')
        assert(out['name'] == 'a')
        state = out['state']
        assert(len(state) > 8)
        assert(state[2] == True)
        assert_allclose(state[1], 2.0)
        assert_allclose(state[4], 0.0)

    def test_eval_param1(self):
        out = deval(self.param1)
        assert(isParameter(out))
        assert(out.name == 'a')
        assert(out.vary == True)
        assert(out.value == 2.0)
        assert(out.min == 0)

    def test_encode_param2(self):
        out = encode(self.param2)
        assert(isinstance(out, dict))
        assert(out['__class__'] == 'Parameter')
        assert(out['name'] == 'b')
        state = out['state']
        assert(len(state) > 8)
        assert(state[2] == False)
        assert(len(state[3]) > 2)

    def test_eval_param2(self):
        out = deval(self.param2)
        assert(isParameter(out))
        assert(out.name == 'b')
        assert(out.expr.startswith('sqrt(x/2)'))

    def test_encode_group1(self):
        out = encode(self.group1)
        assert(isinstance(out, dict))
        assert(out['__class__'] == 'Group')
        assert(out['label'] == 'a string')
        assert(isinstance(out['x'], dict))
        assert(isinstance(out['opts'], dict))
        assert(out['opts']['__class__'] == 'Dict')
        assert(out['x']['__class__'] == 'Array')
        assert(out['y']['__class__'] == 'Array')
        assert_allclose(out['opts']['a'],  1.0, rtol=1.e-4)
        assert_allclose(out['x']['value'][:3], [0, 0.5, 1.0], rtol=1.e-4)

    def test_eval_group1(self):
        out = deval(self.group1)
        assert(isgroup(out))
        assert(getattr(out, 'label') == 'a string')
        assert(out.opts['a'] == 1)
        assert_allclose(out.a, 1.03, rtol=1.e-2)
        assert_allclose(out.x, np.linspace(0, 10, 21), rtol=1.e-2)

    def test_encode_group2(self):
        out = encode(self.group2)
        assert(isinstance(out, dict))
        assert(out['__class__'] == 'Group')
        assert(isinstance(out['par1'], dict))
        assert(out['sub']['__class__'] == 'Group')
        assert(out['sub']['label'] == 'a label')
        assert(out['sub']['x']['__class__'] == 'Array')
        assert(out['sub']['x']['__class__'] == 'Array')
        assert_allclose(out['sub']['x']['value'][:3], [0, 0.5, 1.0], rtol=1.e-4)

        assert(out['par1']['__class__'] == 'Parameter')
        assert(out['par1']['name'] == 'p1')
        assert(out['par2']['__class__'] == 'Parameter')
        assert(out['par2']['name'] == 'p2')
        state1 = out['par1']['state']
        state2 = out['par2']['state']
        assert(state1[2] == False)
        assert_allclose(state1[1], 3.0)
        assert_allclose(state1[4], 0.0)
        assert(state2[2] == True)
        assert_allclose(state2[1], 1.0)


    def test_eval_group2(self):
        out = deval(self.group2)
        assert(isgroup(out))
        assert(out.sub.label == 'a label')
        assert(isParameter(out.par1))
        assert(isParameter(out.par2))
        assert(out.par1.name == 'p1')
        assert(out.par2.name == 'p2')
        assert(out.par1.vary == False)
        assert(out.par2.vary == True)
        assert(out.par1.value == 3.0)
        assert(out.par2.value == 1.0)
        assert(out.par1.min   == 0.0)

    def test_eval_group3(self):
        out = deval(self.group3)
        assert(isgroup(out))
        assert(out.dx['a']     == 1.3 + 0.2j)
        assert(out.dx['tup']   == (0, None))
        assert(out.dx['blist'] == [1,2,3.0])

if __name__ == '__main__':
    for suite in (EncodeDecode4Json_Test,):
        suite = unittest.TestLoader().loadTestsFromTestCase(suite)
        unittest.TextTestRunner(verbosity=0).run(suite)



import sys
import copy
import time
from collections import namedtuple
from functools import partial
from pathlib import Path
import numpy as np
from lmfit import Parameters, minimize, fit_report
from matplotlib.ticker import FuncFormatter

import wx
from wxmplot import PlotPanel
from xraydb import guess_edge
from larch import Group, isgroup
from larch.math import index_of, index_nearest, interp
from larch.utils.strutils import file2groupname, unique_name
from larch.utils import path_split

from larch.wxlib import (GridPanel, BitmapButton, FloatCtrl, FloatSpin,
                         set_color, FloatSpinWithPin, get_icon, SimpleText,
                         Choice, SetTip, Check, Button, HLine, OkCancel,
                         LEFT, pack, plotlabels, ReportFrame, DictFrame,
                         FileCheckList, Font, FONTSIZE, plotlabels,
                         get_zoomlimits, set_zoomlimits)

from larch.xafs import etok, ktoe, find_energy_step
from larch.utils.physical_constants import PI, DEG2RAD, PLANCK_HC, ATOM_SYMS
from larch.math import smooth

Plot_Choices = {'Normalized': 'norm', 'Derivative': 'dmude'}


EDGE_LIST = ('K', 'L3', 'L2', 'L1', 'M5', 'M4', 'M3')

NORM_MU = 'Normalized \u03BC(E)'

DEGLITCH_PLOTS = {'Raw \u03BC(E)': 'mu',
                  NORM_MU: 'norm',
                  '\u03c7(E)': 'chie',
                  '\u03c7(E)*(E-E_0)': 'chiew'}

SESSION_PLOTS = {'Normalized \u03BC(E)': 'norm',
                 'Raw \u03BC(E)': 'mu',
                 'k^2\u03c7(k)': 'chikw'}


def ensure_en_orig(dgroup):
    if not hasattr(dgroup, 'energy_orig'):
        dgroup.energy_orig = dgroup.energy[:]



def fit_dialog_window(dialog, panel):
    sizer = wx.BoxSizer(wx.VERTICAL)
    sizer.Add(panel, 1, LEFT, 5)
    pack(dialog, sizer)
    dialog.Fit()
    w0, h0 = dialog.GetSize()
    w1, h1 = dialog.GetBestSize()
    dialog.SetSize((max(w0, w1)+25, max(h0, h1)+25))


def get_view_limits(ppanel):
    "get last zoom limits for a plot panel"
    xlim = ppanel.axes.get_xlim()
    ylim = ppanel.axes.get_ylim()
    return (xlim, ylim)

def set_view_limits(ppanel, xlim, ylim):
    "set zoom limits for a plot panel, as found from get_view_limits"
    ppanel.axes.set_xlim(xlim, emit=True)
    ppanel.axes.set_ylim(ylim, emit=True)


class OverAbsorptionDialog(wx.Dialog):
    """dialog for correcting over-absorption"""
    def __init__(self, parent, controller, **kws):
        self.parent = parent
        self.controller = controller
        self.dgroup = self.controller.get_group()
        groupnames = list(self.controller.file_groups.keys())

        self.data = [self.dgroup.energy[:], self.dgroup.norm[:]]

        wx.Dialog.__init__(self, parent, wx.ID_ANY, size=(550, 400),
                           title="Correct Over-absorption")
        self.SetFont(Font(FONTSIZE))
        panel = GridPanel(self, ncols=3, nrows=4, pad=4, itemstyle=LEFT)
        self.wids = wids = {}

        wids['grouplist'] = Choice(panel, choices=groupnames, size=(250, -1),
                                   action=self.on_groupchoice)

        wids['grouplist'].SetStringSelection(self.dgroup.filename)

        opts  = dict(size=(90, -1), precision=1, act_on_losefocus=True,
                     minval=-90, maxval=180)

        fs_opts = dict(size=(90, -1), value=45, digits=1, increment=1)
        wids['phi_in']  = FloatSpin(panel, **fs_opts)
        wids['phi_out'] = FloatSpin(panel, **fs_opts)

        wids['elem'] = Choice(panel, choices=ATOM_SYMS[:98], size=(50, -1))
        wids['edge'] = Choice(panel, choices=EDGE_LIST, size=(50, -1))

        wids['formula'] = wx.TextCtrl(panel, -1, '', size=(250, -1))

        self.set_default_elem_edge(self.dgroup)

        #wids['apply'] = Button(panel, 'Save / Overwrite', size=(150, -1),
        #                       action=self.on_apply)
        #SetTip(wids['apply'], 'Save corrected data, overwrite current arrays')

        wids['save_as'] = Button(panel, 'Save As New Group: ', size=(150, -1),
                           action=self.on_saveas)
        SetTip(wids['save_as'], 'Save corrected data as new group')

        wids['save_as_name'] = wx.TextCtrl(panel, -1, self.dgroup.filename + '_abscorr',
                                           size=(250, -1))
        wids['correct'] = Button(panel, 'Do Correction',
                                 size=(150, -1), action=self.on_correct)
        SetTip(wids['correct'], 'Calculate Correction')

        def add_text(text, dcol=1, newrow=True):
            panel.Add(SimpleText(panel, text), dcol=dcol, newrow=newrow)

        add_text(' Correction for Group: ', newrow=False)
        panel.Add(wids['grouplist'], dcol=5)

        add_text(' Absorbing Element: ')
        panel.Add(wids['elem'])

        add_text('  Edge:  ', newrow=False)
        panel.Add(wids['edge'])

        add_text(' Material Formula: ')
        panel.Add(wids['formula'], dcol=3)

        add_text(' Incident Angle (deg): ')
        panel.Add(wids['phi_in'])

        add_text(' Exit Angle (deg): ')
        panel.Add(wids['phi_out'])

        panel.Add(wids['correct'], newrow=True)
        # panel.Add(wids['apply'], dcol=2, newrow=True)

        panel.Add(wids['save_as'], newrow=True)
        panel.Add(wids['save_as_name'], dcol=3)
        panel.Add(Button(panel, 'Done', size=(150, -1), action=self.onDone),
                  newrow=True)
        panel.pack()
        fit_dialog_window(self, panel)
        self.plot_results(keep_limits=False)


    def onDone(self, event=None):
        self.Destroy()

    def set_default_elem_edge(self, dgroup):
        elem, edge = guess_edge(dgroup.e0)
        self.wids['elem'].SetStringSelection(elem)
        self.wids['edge'].SetStringSelection(edge)

    def on_groupchoice(self, event=None):
        fname = self.wids['grouplist'].GetStringSelection()
        self.dgroup = self.controller.get_group(fname)
        self.set_default_elem_edge(self.dgroup)
        self.wids['save_as_name'].SetValue(self.dgroup.filename + '_abscorr')


    def on_correct(self, event=None):
        wids = self.wids
        dgroup = self.dgroup
        anginp = wids['phi_in'].GetValue()
        angout = wids['phi_out'].GetValue()
        elem   = wids['elem'].GetStringSelection()
        edge   = wids['edge'].GetStringSelection()
        formula = wids['formula'].GetValue()
        if len(formula) < 1:
            return

        cmd = """fluo_corr(%s.energy, %s.mu, '%s', '%s', edge='%s', group=%s,
     anginp=%.1f, angout=%.1f)""" % (dgroup.groupname, dgroup.groupname,
                                     formula, elem, edge, dgroup.groupname,
                                     anginp, angout)
        self.cmd = cmd
        self.controller.larch.eval(cmd)
        self.plot_results()

    def on_apply(self, event=None):
        xplot, yplot = self.data
        dgroup = self.dgroup
        dgroup.xplot = dgroup.energy = xplot
        self.parent.process_normalization(dgroup)
        dgroup.journal.add('fluor_corr_command', self.cmd)
        self.plot_results()

    def on_saveas(self, event=None):
        wids = self.wids
        fname = self.wids['grouplist'].GetStringSelection()
        new_fname = wids['save_as_name'].GetValue()
        ngroup = self.controller.copy_group(fname, new_filename=new_fname)

        if hasattr(self.dgroup, 'norm_corr' ):
            ngroup.mu = ngroup.norm_corr*1.0
            del ngroup.norm_corr

        ogroup = self.controller.get_group(fname)
        self.parent.install_group(ngroup, journal=ogroup.journal)
        olddesc = ogroup.journal.get('source_desc').value
        ngroup.journal.add('source_desc', f"fluo_corrected({olddesc})")
        ngroup.journal.add('fluor_correction_command', self.cmd)

    def plot_results(self, event=None, keep_limits=True):
        ppanel = self.controller.get_display(stacked=False).panel

        dgroup = self.dgroup
        xlim, ylim = get_view_limits(ppanel)
        path, fname = path_split(dgroup.filename)

        opts = dict(linewidth=3, ylabel=plotlabels.norm,
                    xlabel=plotlabels.energy, delay_draw=True,
                    show_legend=True)

        if self.controller.plot_erange is not None:
            opts['xmin'] = dgroup.e0 + self.controller.plot_erange[0]
            opts['xmax'] = dgroup.e0 + self.controller.plot_erange[1]

        if not hasattr(dgroup, 'norm_corr'):
            dgroup.norm_corr = dgroup.norm[:]

        ppanel.plot(dgroup.energy, dgroup.norm_corr, zorder=10, marker=None,
                    title='Over-absorption Correction:\n %s' % fname,
                    label='corrected', **opts)

        ppanel.oplot(dgroup.energy, dgroup.norm, zorder=10, marker='o',
                     markersize=3, label='original', **opts)
        if keep_limits:
            set_view_limits(ppanel, xlim, ylim)
        ppanel.canvas.draw()
        ppanel.conf.draw_legend(show=True)

    def GetResponse(self):
        raise AttributeError("use as non-modal dialog!")


class EnergyCalibrateDialog(wx.Dialog):
    """dialog for calibrating energy"""
    def __init__(self, parent, controller, **kws):

        self.parent = parent
        self.controller = controller
        self.dgroup = self.controller.get_group()
        groupnames = list(self.controller.file_groups.keys())

        ensure_en_orig(self.dgroup)

        self.data = [self.dgroup.energy_orig[:], self.dgroup.norm[:]]
        xmin = min(self.dgroup.energy_orig)
        xmax = max(self.dgroup.energy_orig)
        e0val = getattr(self.dgroup, 'e0', xmin)

        wx.Dialog.__init__(self, parent, wx.ID_ANY, size=(550, 400),
                           title="Calibrate / Align Energy")
        self.SetFont(Font(FONTSIZE))
        panel = GridPanel(self, ncols=3, nrows=4, pad=4, itemstyle=LEFT)

        self.wids = wids = {}
        wids['grouplist'] = Choice(panel, choices=groupnames, size=(275, -1),
                                   action=self.on_groupchoice)
        wids['grouplist'].SetStringSelection(self.dgroup.filename)

        refgroups = ['None'] + groupnames
        wids['reflist'] = Choice(panel, choices=refgroups, size=(275, -1),
                                 action=self.on_align, default=0)

        opts = dict(size=(90, -1), digits=3, increment=0.1)

        opts['action'] = partial(self.on_calib, name='eshift')
        wids['eshift'] = FloatSpin(panel, value=0, **opts)

        self.plottype = Choice(panel, choices=list(Plot_Choices.keys()),
                                   size=(275, -1), action=self.plot_results)
        wids['do_align'] = Button(panel, 'Auto Align', size=(100, -1),
                                  action=self.on_align)

        wids['apply_one'] = Button(panel, 'Apply to Current Group', size=(200, -1),
                                   action=self.on_apply_one)

        wids['apply_sel'] = Button(panel, 'Apply to Selected Groups',
                                   size=(250, -1),  action=self.on_apply_sel)
        SetTip(wids['apply_sel'], 'Apply the Energy Shift to all Selected Groups')

        wids['save_as'] = Button(panel, 'Save As New Group ', size=(200, -1),
                           action=self.on_saveas)
        SetTip(wids['save_as'], 'Save shifted data as new group')

        wids['save_as_name'] = wx.TextCtrl(panel, -1,
                                           self.dgroup.filename + '_eshift',
                                           size=(275, -1))

        wids['sharedref_msg'] = wx.StaticText(panel, label="1 groups share this energy reference")
        wids['select_sharedref'] = Button(panel, 'Select Groups with shared reference',
                                          size=(300, -1),  action=self.on_select_sharedrefs)

        def add_text(text, dcol=1, newrow=True):
            panel.Add(SimpleText(panel, text), dcol=dcol, newrow=newrow)

        add_text(' Current Group: ',  newrow=False)
        panel.Add(wids['grouplist'], dcol=2)

        add_text(' Auto-Align to : ')
        panel.Add(wids['reflist'], dcol=2)

        add_text(' Plot Arrays as: ')
        panel.Add(self.plottype, dcol=2)

        add_text(' Energy Shift (eV): ')
        panel.Add(wids['eshift'], dcol=1)
        panel.Add(wids['do_align'], dcol=1)
        panel.Add(HLine(panel, size=(500, 3)), dcol=4, newrow=True)
        # panel.Add(apply_one, newrow=True)

        panel.Add(wids['sharedref_msg'], dcol=2, newrow=True)
        panel.Add(wids['select_sharedref'], dcol=2)
        panel.Add(wids['apply_one'], dcol=1, newrow=True)
        panel.Add(wids['apply_sel'], dcol=2)

        panel.Add(HLine(panel, size=(500, 3)), dcol=4, newrow=True)

        panel.Add(wids['save_as'], newrow=True)
        panel.Add(wids['save_as_name'], dcol=3)

        panel.pack()

        fit_dialog_window(self, panel)

        self.plot_results(keep_limits=False)
        wx.CallAfter(self.get_groups_shared_energyrefs)

    def on_select(self, event=None, opt=None):
        _x, _y = self.controller.get_cursor()
        if opt in self.wids:
            self.wids[opt].SetValue(_x)

    def get_groups_shared_energyrefs(self, dgroup=None):
        if dgroup is None:
            dgroup = self.controller.get_group(self.wids['grouplist'].GetStringSelection())
        sharedrefs = [dgroup.filename]
        try:
            eref = dgroup.config.xasnorm.get('energy_ref', None)
        except:
            eref = None
        if eref is None:
            eref = dgroup.groupname


        for key, val in self.controller.file_groups.items():
            if dgroup.groupname == val:
                continue
            g = self.controller.get_group(val)
            try:
                geref = g.config.xasnorm.get('energy_ref', None)
            except:
                geref = None
            # print(key, val, geref, geref == ref_filename)
            if geref == eref or geref == dgroup.filename:
                sharedrefs.append(key)
        self.wids['sharedref_msg'].SetLabel(f" {len(sharedrefs):d} groups share this energy reference")
        return sharedrefs

    def on_select_sharedrefs(self, event=None):
        groups = self.get_groups_shared_energyrefs()
        self.controller.filelist.SetCheckedStrings(groups)

    def on_groupchoice(self, event=None):
        dgroup = self.controller.get_group(self.wids['grouplist'].GetStringSelection())
        self.dgroup = dgroup
        others = self.get_groups_shared_energyrefs(dgroup)
        self.wids['save_as_name'].SetValue(self.dgroup.filename + '_eshift')
        self.plot_results()

    def on_align(self, event=None, name=None, value=None):
        ref = self.controller.get_group(self.wids['reflist'].GetStringSelection())
        dat = self.dgroup
        ensure_en_orig(dat)
        ensure_en_orig(ref)

        dat.xplot = dat.energy_orig[:]
        ref.xplot = ref.energy_orig[:]
        estep = find_energy_step(dat.xplot)
        i1 = index_of(ref.energy_orig, ref.e0-20)
        i2 = index_of(ref.energy_orig, ref.e0+20)

        def resid(pars, ref, dat, i1, i2):
            "fit residual"
            newx = dat.xplot + pars['eshift'].value
            scale = pars['scale'].value
            y = interp(newx, dat.dmude, ref.xplot, kind='cubic')
            return smooth(ref.xplot, y*scale-ref.dmude, xstep=estep, sigma=0.50)[i1:i2]

        params = Parameters()
        ex0 = ref.e0-dat.e0
        emax = 50.0
        if abs(ex0) > 75:
            ex0 = 0.00
            emax = (abs(ex0) + 75.0)
        elif abs(ex0) > 10:
            emax = (abs(ex0) + 75.0)
        params.add('eshift', value=ex0, min=-emax, max=emax)
        params.add('scale', value=1, min=1.e-8, max=50)
        # print("Fit params ", params)
        result = minimize(resid, params, args=(ref, dat, i1, i2),
                          max_nfev=1000)
        # print(fit_report(result))
        eshift = result.params['eshift'].value
        self.wids['eshift'].SetValue(eshift)

        ensure_en_orig(self.dgroup)
        xnew = self.dgroup.energy_orig + eshift
        self.data = xnew, self.dgroup.norm[:]
        self.plot_results()

    def on_calib(self, event=None, name=None):
        wids = self.wids
        eshift = wids['eshift'].GetValue()
        ensure_en_orig(self.dgroup)
        xnew = self.dgroup.energy_orig + eshift
        self.data = xnew, self.dgroup.norm[:]
        self.plot_results()

    def on_apply_one(self, event=None):
        xplot, yplot = self.data
        dgroup = self.dgroup
        eshift = self.wids['eshift'].GetValue()

        ensure_en_orig(dgroup)

        idx, norm_page = self.parent.get_nbpage('xasnorm')
        norm_page.wids['energy_shift'].SetValue(eshift)

        dgroup.energy_shift = eshift
        dgroup.xplot = dgroup.energy = eshift + dgroup.energy_orig[:]
        dgroup.journal.add('energy_shift ', eshift)
        self.parent.process_normalization(dgroup)
        self.plot_results()

    def on_apply_sel(self, event=None):
        eshift = self.wids['eshift'].GetValue()
        idx, norm_page = self.parent.get_nbpage('xasnorm')
        for checked in self.controller.filelist.GetCheckedStrings():
            fname  = self.controller.file_groups[str(checked)]
            dgroup = self.controller.get_group(fname)
            ensure_en_orig(dgroup)
            dgroup.energy_shift = eshift
            norm_page.wids['energy_shift'].SetValue(eshift)

            dgroup.xplot = dgroup.energy = eshift + dgroup.energy_orig[:]
            dgroup.journal.add('energy_shift ', eshift)
            self.parent.process_normalization(dgroup)

    def on_saveas(self, event=None):
        wids = self.wids
        fname = wids['grouplist'].GetStringSelection()
        eshift = wids['eshift'].GetValue()
        new_fname = wids['save_as_name'].GetValue()
        ngroup = self.controller.copy_group(fname, new_filename=new_fname)

        ensure_en_orig(ngroup)
        ngroup.xplot = ngroup.energy = eshift + ngroup.energy_orig[:]
        ngroup.energy_shift = 0
        ngroup.energy_ref = ngroup.groupname

        ogroup = self.controller.get_group(fname)
        self.parent.install_group(ngroup, journal=ogroup.journal)
        olddesc = ogroup.journal.get('source_desc').value
        ngroup.journal.add('source_desc', f"energy_shifted({olddesc}, {eshift:.4f})")
        ngroup.journal.add('energy_shift ', 0.0)

    def plot_results(self, event=None, keep_limits=True):
        ppanel = self.controller.get_display(stacked=False).panel
        ppanel.oplot
        xnew, ynew = self.data
        dgroup = self.dgroup

        xlim, ylim = get_view_limits(ppanel)
        path, fname = path_split(dgroup.filename)

        wids = self.wids
        eshift = wids['eshift'].GetValue()
        e0_old = dgroup.e0
        e0_new = dgroup.e0 + eshift

        xmin = min(e0_old, e0_new) - 25
        xmax = max(e0_old, e0_new) + 50

        use_deriv = self.plottype.GetStringSelection().lower().startswith('deriv')

        ylabel = plotlabels.norm
        if use_deriv:
            ynew = np.gradient(ynew)/np.gradient(xnew)
            ylabel = plotlabels.dmude

        opts = dict(xmin=xmin, xmax=xmax, ylabel=ylabel, delay_draw=True,
                    xlabel=plotlabels.energy, show_legend=True)

        if self.controller.plot_erange is not None:
            opts['xmin'] = dgroup.e0 + self.controller.plot_erange[0]
            opts['xmax'] = dgroup.e0 + self.controller.plot_erange[1]

        xold, yold = self.dgroup.energy_orig, self.dgroup.norm
        if use_deriv:
            yold = np.gradient(yold)/np.gradient(xold)

        ppanel.plot(xold, yold, zorder=10, marker='o', markersize=3,
                     label='original', linewidth=2, color='#1f77b4',
                     title=f'Energy Calibration:\n {fname}', **opts)

        ppanel.oplot(xnew, ynew, zorder=15, marker='+', markersize=3,
                    linewidth=2, label='shifted',
                    color='#d62728', **opts)

        if wids['reflist'].GetStringSelection() != 'None':
            refgroup = self.controller.get_group(wids['reflist'].GetStringSelection())
            xref, yref = refgroup.energy, refgroup.norm
            if use_deriv:
                yref = np.gradient(yref)/np.gradient(xref)

            ppanel.oplot(xref, yref, style='solid', zorder=5, color='#2ca02c',
                         marker=None, label=refgroup.filename, **opts)
        if keep_limits:
            set_view_limits(ppanel, xlim, ylim)
        ppanel.canvas.draw()


    def GetResponse(self):
        raise AttributeError("use as non-modal dialog!")

class RebinDataDialog(wx.Dialog):
    """dialog for rebinning data to standard XAFS grid"""
    def __init__(self, parent, controller, **kws):

        self.parent = parent
        self.controller = controller
        self.dgroup = self.controller.get_group()
        groupnames = list(self.controller.file_groups.keys())

        xmin = min(self.dgroup.energy)
        xmax = max(self.dgroup.energy)
        e0val = getattr(self.dgroup, 'e0', xmin)
        xanes_step = 0.05 * (1 + max(1, int(e0val / 2000.0)))
        self.data = [self.dgroup.energy[:], self.dgroup.mu[:],
                     self.dgroup.mu*0, e0val]

        wx.Dialog.__init__(self, parent, wx.ID_ANY, size=(550, 400),
                           title="Rebin mu(E) Data")
        self.SetFont(Font(FONTSIZE))
        panel = GridPanel(self, ncols=3, nrows=4, pad=2, itemstyle=LEFT)

        self.wids = wids = {}

        wids['grouplist'] = Choice(panel, choices=groupnames, size=(250, -1),
                                action=self.on_groupchoice)

        wids['grouplist'].SetStringSelection(self.dgroup.groupname)

        opts  = dict(size=(90, -1), precision=3, act_on_losefocus=True)

        wids['e0'] = FloatCtrl(panel, value=e0val, minval=xmin, maxval=xmax, **opts)
        pre1 = 10.0*(1+int((xmin-e0val)/10.0))
        wids['pre1'] = FloatCtrl(panel, value=pre1,  **opts)
        wids['pre2'] = FloatCtrl(panel, value=-15, **opts)
        wids['xanes1'] = FloatCtrl(panel, value=-15,  **opts)
        wids['xanes2'] = FloatCtrl(panel, value=15, **opts)
        wids['exafs1'] = FloatCtrl(panel, value=etok(15),  **opts)
        wids['exafs2'] = FloatCtrl(panel, value=etok(xmax-e0val), **opts)

        wids['pre_step'] = FloatCtrl(panel, value=2.0,  **opts)
        wids['xanes_step'] = FloatCtrl(panel, value=xanes_step,  **opts)
        wids['exafs_step'] = FloatCtrl(panel, value=0.05,  **opts)

        for wname, wid in wids.items():
            if wname != 'grouplist':
                wid.SetAction(partial(self.on_rebin, name=wname))

        #wids['apply'] = Button(panel, 'Save / Overwrite', size=(150, -1),
        #                       action=self.on_apply)
        #SetTip(wids['apply'], 'Save rebinned data, overwrite current arrays')

        wids['save_as'] = Button(panel, 'Save As New Group: ', size=(150, -1),
                                 action=self.on_saveas)
        SetTip(wids['save_as'], 'Save corrected data as new group')

        wids['save_as_name'] = wx.TextCtrl(panel, -1, self.dgroup.filename + '_rebin',
                                           size=(250, -1))

        def add_text(text, dcol=1, newrow=True):
            panel.Add(SimpleText(panel, text), dcol=dcol, newrow=newrow)

        add_text('Rebin Data for Group: ', dcol=2, newrow=False)
        panel.Add(wids['grouplist'], dcol=3)

        add_text('E0: ')
        panel.Add(wids['e0'])
        add_text(' eV', newrow=False)

        add_text('Region ')
        add_text('Start ', newrow=False)
        add_text('Stop ', newrow=False)
        add_text('Step ', newrow=False)
        add_text('Units ', newrow=False)

        add_text('Pre-Edge: ')
        panel.Add(wids['pre1'])
        panel.Add(wids['pre2'])
        panel.Add(wids['pre_step'])
        add_text(' eV', newrow=False)

        add_text('XANES: ')
        panel.Add(wids['xanes1'])
        panel.Add(wids['xanes2'])
        panel.Add(wids['xanes_step'])
        add_text(' eV', newrow=False)

        add_text('EXAFS: ')
        panel.Add(wids['exafs1'])
        panel.Add(wids['exafs2'])
        panel.Add(wids['exafs_step'])
        add_text('1/\u212B', newrow=False)

        # panel.Add(wids['apply'], dcol=2, newrow=True)
        panel.Add(wids['save_as'],  dcol=2, newrow=True)
        panel.Add(wids['save_as_name'], dcol=3)
        panel.Add(Button(panel, 'Done', size=(150, -1), action=self.onDone),
                  newrow=True)
        panel.pack()

        fit_dialog_window(self, panel)

        self.on_rebin()
        self.plot_results(keep_limits=False)

    def onDone(self, event=None):
        self.Destroy()

    def on_groupchoice(self, event=None):
        self.dgroup = self.controller.get_group(self.wids['grouplist'].GetStringSelection())
        self.wids['save_as_name'].SetValue(self.dgroup.filename + '_rebin')
        self.plot_results()

    def on_rebin(self, event=None, name=None, value=None):
        wids = self.wids
        if name == 'pre2':
            val = wids['pre2'].GetValue()
            wids['xanes1'].SetValue(val, act=False)
        elif name == 'xanes1':
            val = wids['xanes1'].GetValue()
            wids['pre2'].SetValue(val, act=False)
        elif name == 'xanes2':
            val = wids['xanes2'].GetValue()
            wids['exafs1'].SetValue(etok(val), act=False)
        elif name == 'exafs1':
            val = wids['exafs1'].GetValue()
            wids['xanes2'].SetValue(ktoe(val), act=False)

        e0 = wids['e0'].GetValue()
        args = dict(group=self.dgroup.groupname, e0=e0,
                    pre1=wids['pre1'].GetValue(),
                    pre2=wids['pre2'].GetValue(),
                    pre_step=wids['pre_step'].GetValue(),
                    exafs1=ktoe(wids['exafs1'].GetValue()),
                    exafs2=ktoe(wids['exafs2'].GetValue()),
                    exafs_kstep=wids['exafs_step'].GetValue(),
                    xanes_step=wids['xanes_step'].GetValue())

        # do rebin:
        cmd = """rebin_xafs({group}, e0={e0:f}, pre1={pre1:f}, pre2={pre2:f},
        pre_step={pre_step:f}, xanes_step={xanes_step:f}, exafs1={exafs1:f},
        exafs2={exafs2:f}, exafs_kstep={exafs_kstep:f})""".format(**args)
        self.cmd = cmd
        self.controller.larch.eval(cmd)

        if hasattr(self.dgroup, 'rebinned'):
            xnew = self.dgroup.rebinned.energy
            ynew = self.dgroup.rebinned.mu
            yerr = self.dgroup.rebinned.delta_mu
            self.data = xnew, ynew, yerr, e0
            self.plot_results()

    def on_apply(self, event=None):
        xplot, yplot, yerr, e0 = self.data
        dgroup = self.dgroup
        dgroup.energy = dgroup.xplot = xplot
        dgroup.mu     = dgroup.yplot = yplot
        dgroup.journal.add('rebin_command ', self.cmd)
        self.parent.process_normalization(dgroup)
        self.plot_results()

    def on_saveas(self, event=None):
        wids = self.wids
        fname = wids['grouplist'].GetStringSelection()
        new_fname = wids['save_as_name'].GetValue()
        ngroup = self.controller.copy_group(fname, new_filename=new_fname)
        xplot, yplot, yerr, de0 = self.data
        ngroup.energy = ngroup.xplot = xplot
        ngroup.mu     = ngroup.yplot = yplot

        ogroup = self.controller.get_group(fname)
        olddesc = ogroup.journal.get('source_desc').value

        ngroup.delta_mu = getattr(ngroup, 'yerr', 1.0)
        self.parent.process_normalization(ngroup)

        self.parent.install_group(ngroup, journal=ogroup.journal)
        ngroup.journal.add('source_desc', f"rebinned({olddesc})")
        ngroup.journal.add('rebin_command ', self.cmd)

    def on_done(self, event=None):
        self.Destroy()

    def plot_results(self, event=None, keep_limits=True):
        ppanel = self.controller.get_display(stacked=False).panel
        xnew, ynew, yerr, e0 = self.data
        dgroup = self.dgroup
        xlim, ylim = get_view_limits(ppanel)
        path, fname = path_split(dgroup.filename)

        opts = {'delay_draw': True}
        if self.controller.plot_erange is not None:
            opts['xmin'] = dgroup.e0 + self.controller.plot_erange[0]
            opts['xmax'] = dgroup.e0 + self.controller.plot_erange[1]

        ppanel.plot(xnew, ynew, zorder=20, marker='square',
                    linewidth=3, title='Enegy rebinning:\n %s' % fname,
                    label='rebinned', xlabel=plotlabels.energy,
                    ylabel=plotlabels.mu, **opts)

        xold, yold = self.dgroup.energy, self.dgroup.mu
        ppanel.oplot(xold, yold, zorder=10,
                     marker='o', markersize=4, linewidth=2.0,
                     label='original', show_legend=True, **opts)
        if keep_limits:
            set_view_limits(ppanel, xlim, ylim)
        ppanel.canvas.draw()

    def GetResponse(self):
        raise AttributeError("use as non-modal dialog!")

class SmoothDataDialog(wx.Dialog):
    """dialog for smoothing data"""
    def __init__(self, parent, controller, **kws):

        self.parent = parent
        self.controller = controller
        self.dgroup = self.controller.get_group()
        groupnames = list(self.controller.file_groups.keys())

        self.data = [self.dgroup.energy[:], self.dgroup.mu[:]]


        wx.Dialog.__init__(self, parent, wx.ID_ANY, size=(550, 400),
                           title="Smooth mu(E) Data")
        self.SetFont(Font(FONTSIZE))
        panel = GridPanel(self, ncols=3, nrows=4, pad=4, itemstyle=LEFT)

        self.wids = wids = {}

        wids['grouplist'] = Choice(panel, choices=groupnames, size=(250, -1),
                                action=self.on_groupchoice)

        wids['grouplist'].SetStringSelection(self.dgroup.filename)
        SetTip(wids['grouplist'], 'select a new group, clear undo history')

        smooth_ops = ('None', 'Boxcar', 'Savitzky-Golay', 'Convolution')
        conv_ops  = ('Lorenztian', 'Gaussian')

        self.smooth_op = Choice(panel, choices=smooth_ops, size=(150, -1),
                                action=self.on_smooth)
        self.smooth_op.SetSelection(0)

        self.conv_op = Choice(panel, choices=conv_ops, size=(150, -1),
                                action=self.on_smooth)
        self.conv_op.SetSelection(0)

        opts  = dict(size=(50, -1), act_on_losefocus=True, odd_only=False)

        self.sigma = FloatSpin(panel, value=1, digits=2, min_val=0, increment=0.1,
                               size=(60, -1), action=self.on_smooth)

        self.par_n = FloatSpin(panel, value=2, digits=0, min_val=1, increment=1,
                               size=(60, -1), action=self.on_smooth)

        self.par_o = FloatSpin(panel, value=1, digits=0, min_val=1, increment=1,
                               size=(60, -1), action=self.on_smooth)

        # for fc in (self.sigma, self.par_n, self.par_o):
        # self.sigma.SetAction(self.on_smooth)

        self.message = SimpleText(panel, label='         ', size=(200, -1))


        wids['save_as'] = Button(panel, 'Save As New Group: ', size=(150, -1),
                           action=self.on_saveas)
        SetTip(wids['save_as'], 'Save corrected data as new group')

        wids['save_as_name'] = wx.TextCtrl(panel, -1, self.dgroup.filename + '_smooth',
                                           size=(250, -1))

        def add_text(text, dcol=1, newrow=True):
            panel.Add(SimpleText(panel, text), dcol=dcol, newrow=newrow)

        add_text('Smooth Data for Group: ', newrow=False)
        panel.Add(wids['grouplist'], dcol=5)

        add_text('Smoothing Method: ')
        panel.Add(self.smooth_op)
        add_text(' n = ', newrow=False)
        panel.Add(self.par_n)
        add_text(' order=  ', newrow=False)
        panel.Add(self.par_o)

        add_text('Convolution Form: ')
        panel.Add(self.conv_op)
        add_text(' sigma: ', newrow=False)
        panel.Add(self.sigma)

        panel.Add((10, 10), newrow=True)
        panel.Add(self.message, dcol=5)

        # panel.Add(wids['apply'], newrow=True)

        panel.Add(wids['save_as'],  newrow=True)
        panel.Add(wids['save_as_name'], dcol=5)
        panel.Add(Button(panel, 'Done', size=(150, -1), action=self.onDone),
                  newrow=True)
        panel.pack()
        fit_dialog_window(self, panel)

        self.plot_results(keep_limits=False)

    def onDone(self, event=None):
        self.Destroy()


    def on_groupchoice(self, event=None):
        self.dgroup = self.controller.get_group(self.wids['grouplist'].GetStringSelection())
        self.wids['save_as_name'].SetValue(self.dgroup.filename + '_smooth')
        self.plot_results()

    def on_smooth(self, event=None, value=None):
        smoothop = self.smooth_op.GetStringSelection().lower()

        convop   = self.conv_op.GetStringSelection()
        self.conv_op.Enable(smoothop.startswith('conv'))
        self.sigma.Enable(smoothop.startswith('conv'))
        self.message.SetLabel('')
        self.par_n.SetMin(1)
        self.par_n.odd_only = False
        par_n = int(self.par_n.GetValue())
        par_o = int(self.par_o.GetValue())
        sigma = self.sigma.GetValue()
        cmd = '{group:s}.mu' # No smoothing
        estep = find_energy_step(self.data[0])
        if smoothop.startswith('box'):
            self.par_n.Enable()
            cmd = "boxcar({group:s}.mu, {par_n:d})"
            self.conv_op.Disable()
        elif smoothop.startswith('savi'):
            self.par_n.Enable()
            self.par_n.odd_only = True
            self.par_o.Enable()

            x0 = max(par_o + 1, par_n)
            if x0 % 2 == 0:
                x0 += 1
            self.par_n.SetMin(par_o + 1)
            if par_n != x0:
                self.par_n.SetValue(x0)
            self.message.SetLabel('n must odd and > order+1')

            cmd = "savitzky_golay({group:s}.mu, {par_n:d}, {par_o:d})"

        elif smoothop.startswith('conv'):
            cmd = "smooth({group:s}.energy, {group:s}.mu, xstep={estep:f}, sigma={sigma:f}, form='{convop:s}')"
        self.cmd = cmd.format(group=self.dgroup.groupname, convop=convop,
                              estep=estep, sigma=sigma, par_n=par_n, par_o=par_o)

        self.controller.larch.eval("_tmpy = %s" % self.cmd)
        self.data = self.dgroup.energy[:], self.controller.symtable._tmpy
        self.plot_results()

    def on_apply(self, event=None):
        xplot, yplot = self.data
        dgroup = self.dgroup
        dgroup.energy = xplot
        dgroup.mu     = yplot
        dgroup.journal.add('smooth_command', self.cmd)
        self.parent.process_normalization(dgroup)
        self.plot_results()

    def on_saveas(self, event=None):
        wids = self.wids
        fname = wids['grouplist'].GetStringSelection()
        new_fname = wids['save_as_name'].GetValue()
        ngroup = self.controller.copy_group(fname, new_filename=new_fname)

        xplot, yplot = self.data
        ngroup.energy = ngroup.xplot = xplot
        ngroup.mu     = ngroup.yplot = yplot

        ogroup = self.controller.get_group(fname)
        olddesc = ogroup.journal.get('source_desc').value

        self.parent.install_group(ngroup, journal=ogroup.journal)
        ngroup.journal.add('source_desc', f"smoothed({olddesc})")
        ngroup.journal.add('smooth_command', self.cmd)
        self.parent.process_normalization(ngroup)

    def on_done(self, event=None):
        self.Destroy()

    def plot_results(self, event=None, keep_limits=True):
        ppanel = self.controller.get_display(stacked=False).panel
        xnew, ynew = self.data
        dgroup = self.dgroup
        path, fname = path_split(dgroup.filename)
        opts = {'delay_draw': True}
        xlim, ylim = get_view_limits(ppanel)

        if self.controller.plot_erange is not None:
            opts['xmin'] = dgroup.e0 + self.controller.plot_erange[0]
            opts['xmax'] = dgroup.e0 + self.controller.plot_erange[1]

        ppanel.plot(xnew, ynew, zorder=20, marker=None,
                    linewidth=3, title='Smoothing:\n %s' % fname,
                    label='smoothed', xlabel=plotlabels.energy,
                    ylabel=plotlabels.mu, **opts)

        xold, yold = self.dgroup.energy, self.dgroup.mu
        ppanel.oplot(xold, yold, zorder=10,
                     marker='o', markersize=4, linewidth=2.0,
                     label='original', show_legend=True, **opts)
        if keep_limits:
            set_view_limits(ppanel, xlim, ylim)
        ppanel.canvas.draw()

    def GetResponse(self):
        raise AttributeError("use as non-modal dialog!")

class DeconvolutionDialog(wx.Dialog):
    """dialog for energy deconvolution"""
    def __init__(self, parent, controller, **kws):

        self.parent = parent
        self.controller = controller
        self.dgroup = self.controller.get_group()
        groupnames = list(self.controller.file_groups.keys())

        self.data = [self.dgroup.energy[:], self.dgroup.norm[:]]


        wx.Dialog.__init__(self, parent, wx.ID_ANY, size=(550, 400),
                           title="Deconvolve mu(E) Data")
        self.SetFont(Font(FONTSIZE))
        panel = GridPanel(self, ncols=3, nrows=4, pad=4, itemstyle=LEFT)

        self.wids = wids = {}

        wids['grouplist'] = Choice(panel, choices=groupnames, size=(250, -1),
                                action=self.on_groupchoice)

        wids['grouplist'].SetStringSelection(self.dgroup.groupname)
        SetTip(wids['grouplist'], 'select a new group, clear undo history')

        deconv_ops  = ('Lorenztian', 'Gaussian')

        wids['deconv_op'] = Choice(panel, choices=deconv_ops, size=(150, -1),
                                   action=self.on_deconvolve)

        wids['esigma'] = FloatSpin(panel, value=0.5, digits=2, size=(90, -1),
                                   increment=0.1, action=self.on_deconvolve)

        #wids['apply'] = Button(panel, 'Save / Overwrite', size=(150, -1),
        #                       action=self.on_apply)
        #SetTip(wids['apply'], 'Save corrected data, overwrite current arrays')

        wids['save_as'] = Button(panel, 'Save As New Group: ', size=(150, -1),
                           action=self.on_saveas)
        SetTip(wids['save_as'], 'Save corrected data as new group')

        wids['save_as_name'] = wx.TextCtrl(panel, -1, self.dgroup.filename + '_deconv',
                                           size=(250, -1))

        def add_text(text, dcol=1, newrow=True):
            panel.Add(SimpleText(panel, text), dcol=dcol, newrow=newrow)

        add_text('Deconvolve Data for Group: ', newrow=False)
        panel.Add(wids['grouplist'], dcol=5)

        add_text('Functional Form: ')
        panel.Add(wids['deconv_op'])

        add_text(' sigma= ')
        panel.Add(wids['esigma'])
        # panel.Add(wids['apply'], newrow=True)
        panel.Add(wids['save_as'],  newrow=True)
        panel.Add(wids['save_as_name'], dcol=5)
        panel.Add(Button(panel, 'Done', size=(150, -1), action=self.onDone),
                  newrow=True)
        panel.pack()

        fit_dialog_window(self, panel)
        self.plot_results(keep_limits=False)

    def onDone(self, event=None):
        self.Destroy()

    def on_saveas(self, event=None):
        wids = self.wids
        fname = wids['grouplist'].GetStringSelection()
        new_fname = wids['save_as_name'].GetValue()
        ngroup = self.controller.copy_group(fname, new_filename=new_fname)
        xplot, yplot = self.data
        ngroup.energy = ngroup.xplot = xplot
        ngroup.mu     = ngroup.yplot = yplot

        ogroup = self.controller.get_group(fname)
        olddesc = ogroup.journal.get('source_desc').value

        self.parent.install_group(ngroup, journal=ogroup.journal)
        ngroup.journal.add('source_desc', f"deconvolved({olddesc})")
        ngroup.journal.add('deconvolve_command', self.cmd)
        self.parent.process_normalization(ngroup)


    def on_groupchoice(self, event=None):
        self.dgroup = self.controller.get_group(self.wids['grouplist'].GetStringSelection())
        self.wids['save_as_name'].SetValue(self.dgroup.filename + '_deconv')
        self.plot_results()

    def on_deconvolve(self, event=None, value=None):
        deconv_form  = self.wids['deconv_op'].GetStringSelection()

        esigma = self.wids['esigma'].GetValue()

        dopts = [self.dgroup.groupname,
                 "form='%s'" % (deconv_form),
                 "esigma=%.4f" % (esigma)]
        self.cmd = "xas_deconvolve(%s)" % (', '.join(dopts))
        self.controller.larch.eval(self.cmd)

        self.data = self.dgroup.energy[:], self.dgroup.deconv[:]
        self.plot_results()

    def on_apply(self, event=None):
        xplot, yplot = self.data
        dgroup = self.dgroup
        dgroup.energy = xplot
        dgroup.mu     = yplot
        dgroup.journal.add('deconvolve_command ', self.cmd)
        self.parent.process_normalization(dgroup)
        self.plot_results()

    def plot_results(self, event=None, keep_limits=True):
        ppanel = self.controller.get_display(stacked=False).panel
        xnew, ynew = self.data
        dgroup = self.dgroup
        xlim, ylim = get_view_limits(ppanel)
        path, fname = path_split(dgroup.filename)

        opts = {'delay_draw': True}
        if self.controller.plot_erange is not None:
            opts['xmin'] = dgroup.e0 + self.controller.plot_erange[0]
            opts['xmax'] = dgroup.e0 + self.controller.plot_erange[1]

        ppanel.plot(xnew, ynew, zorder=20, marker=None,
                    linewidth=3, title='Deconvolving:\n %s' % fname,
                    label='deconvolved', xlabel=plotlabels.energy,
                    ylabel=plotlabels.mu, **opts)

        xold, yold = self.dgroup.energy, self.dgroup.norm
        ppanel.oplot(xold, yold, zorder=10,
                     marker='o', markersize=4, linewidth=2.0,
                     label='original', show_legend=True, **opts)
        if keep_limits:
            set_view_limits(ppanel, xlim, ylim)
        ppanel.canvas.draw()

    def GetResponse(self):
        raise AttributeError("use as non-modal dialog!")

class DeglitchDialog(wx.Dialog):
    """dialog for deglitching or removing unsightly data points"""
    def __init__(self, parent, controller, **kws):
        self.parent = parent
        self.controller = controller
        self.wids = {}
        self.plot_markers = None
        self.dgroup = self.controller.get_group()
        groupnames = list(self.controller.file_groups.keys())

        self.reset_data_history()
        xplot, yplot = self.data

        xrange = (max(xplot) - min(xplot))
        xmax = int(max(xplot) + xrange/5.0)
        xmin = int(min(xplot) - xrange/5.0)

        lastx, lasty = self.controller.get_cursor()
        if lastx is None:
            lastx = max(xplot)

        wx.Dialog.__init__(self, parent, wx.ID_ANY, size=(550, 400),
                           title="Select Points to Remove")
        self.SetFont(Font(FONTSIZE))
        panel = GridPanel(self, ncols=3, nrows=4, pad=4, itemstyle=LEFT)
        wids = self.wids

        wids['grouplist'] = Choice(panel, choices=groupnames, size=(250, -1),
                                action=self.on_groupchoice)

        wids['grouplist'].SetStringSelection(self.dgroup.filename)
        SetTip(wids['grouplist'], 'select a new group, clear undo history')

        br_xlast = Button(panel, 'Remove point', size=(125, -1),
                          action=partial(self.on_remove, opt='x'))

        br_range = Button(panel, 'Remove range', size=(125, -1),
                          action=partial(self.on_remove, opt='range'))

        undo = Button(panel, 'Undo remove', size=(125, -1),
                      action=self.on_undo)

        wids['save_as'] = Button(panel, 'Save As New Group: ', size=(150, -1),
                                 action=self.on_saveas)
        SetTip(wids['save_as'], 'Save deglitched data as new group')

        wids['save_as_name'] = wx.TextCtrl(panel, -1, self.dgroup.filename + '_clean',
                                           size=(250, -1))

        self.history_message = SimpleText(panel, '')

        opts  = dict(size=(125, -1), digits=2, increment=0.1)
        for wname in ('xlast', 'range1', 'range2'):
            if wname == 'range2':
                lastx += 1
            pin_action = partial(self.parent.onSelPoint, opt=wname,
                                 relative_e0=False,
                                 callback=self.on_pinvalue)

            float_action=partial(self.on_floatvalue, opt=wname)
            fspin, pinb = FloatSpinWithPin(panel, value=lastx,
                                           pin_action=pin_action,
                                           action=float_action)
            wids[wname] = fspin
            wids[wname+'_pin'] = pinb

        self.choice_range = Choice(panel, choices=('above', 'below', 'between'),
                                    size=(90, -1), action=self.on_rangechoice)

        self.choice_range.SetStringSelection('above')
        wids['range2'].Disable()

        wids['plotopts'] = Choice(panel, choices=list(DEGLITCH_PLOTS.keys()),
                                  size=(175, -1),
                                  action=self.on_plotchoice)

        wids['plotopts'].SetStringSelection(NORM_MU)

        def add_text(text, dcol=1, newrow=True):
            panel.Add(SimpleText(panel, text), dcol=dcol, newrow=newrow)

        add_text('Deglitch Data for Group: ', dcol=2, newrow=False)
        panel.Add(wids['grouplist'], dcol=5)

        add_text('Single Energy : ', dcol=2)
        panel.Add(wids['xlast'])
        panel.Add(wids['xlast_pin'])
        panel.Add(br_xlast)

        add_text('Plot Data as:  ', dcol=2)
        panel.Add(wids['plotopts'], dcol=5)

        add_text('Energy Range : ')
        panel.Add(self.choice_range)
        panel.Add(wids['range1'])
        panel.Add(wids['range1_pin'])
        panel.Add(br_range)

        panel.Add((10, 10), dcol=2, newrow=True)
        panel.Add(wids['range2'])
        panel.Add(wids['range2_pin'])

        # panel.Add(wids['apply'], dcol=2, newrow=True)

        panel.Add(wids['save_as'], dcol=2, newrow=True)
        panel.Add(wids['save_as_name'], dcol=4)
        panel.Add(Button(panel, 'Done', size=(150, -1), action=self.onDone),
                  dcol=2, newrow=True)
        panel.Add(self.history_message, dcol=2)
        panel.Add(undo)

        panel.pack()

        fit_dialog_window(self, panel)
        self.plot_results(keep_limits=False)

    def onDone(self, event=None):
        self.Destroy()

    def reset_data_history(self):
        plottype = 'norm'
        if 'plotopts' in self.wids:
            plotstr = self.wids['plotopts'].GetStringSelection()
            plottype = DEGLITCH_PLOTS[plotstr]
        self.data = self.get_xydata(datatype=plottype)
        self.xmasks = [np.ones(len(self.data[0]), dtype=bool)]
        self.plot_markers = None

    def get_xydata(self, datatype='mu'):
        if hasattr(self.dgroup, 'energy'):
            xplot = self.dgroup.energy[:]
        else:
            xplot = self.dgroup.xplot[:]
        yplot = self.dgroup.yplot[:]
        if datatype == 'mu' and hasattr(self.dgroup, 'mu'):
            yplot = self.dgroup.mu[:]
        elif datatype == 'norm':
            if not hasattr(self.dgroup, 'norm'):
                self.parent.process_normalization(dgroup)
            yplot = self.dgroup.norm[:]
        elif datatype in ('chie', 'chiew'):
            if not hasattr(self.dgroup, 'chie'):
                self.parent.process_exafs(self.dgroup)
            yplot = self.dgroup.chie[:]
            if datatype == 'chiew':
                yplot = self.dgroup.chie[:] * (xplot-self.dgroup.e0)
        return (xplot, yplot)

    def on_groupchoice(self, event=None):
        self.dgroup = self.controller.get_group(self.wids['grouplist'].GetStringSelection())
        self.wids['save_as_name'].SetValue(self.dgroup.filename + '_clean')
        self.reset_data_history()
        self.plot_results(use_zoom=True)

    def on_rangechoice(self, event=None):
        sel = self.choice_range.GetStringSelection()
        self.wids['range2'].Enable(sel == 'between')


    def on_plotchoice(self, event=None):
        plotstr = self.wids['plotopts'].GetStringSelection()
        plottype = DEGLITCH_PLOTS[plotstr]
        self.data = self.get_xydata(datatype=plottype)
        self.plot_results()

    def on_pinvalue(self, opt='__', xsel=None, **kws):
        if xsel is not None and opt in self.wids:
            self.wids[opt].SetValue(xsel)
        self.plot_markers = opt
        self.plot_results()

    def on_floatvalue(self, val=None, opt='_', **kws):
        self.plot_markers = opt
        self.plot_results()

    def on_remove(self, event=None, opt=None):
        xwork, ywork = self.data
        mask = copy.deepcopy(self.xmasks[-1])
        if opt == 'x':
            bad = index_nearest(xwork, self.wids['xlast'].GetValue())
            mask[bad] = False
        elif opt == 'range':
            rchoice = self.choice_range.GetStringSelection().lower()
            x1 = index_nearest(xwork, self.wids['range1'].GetValue())
            x2 = None
            if rchoice == 'below':
                x2, x1 = x1, x2
            elif rchoice == 'between':
                x2 = index_nearest(xwork, self.wids['range2'].GetValue())
                if x1 > x2:
                    x1, x2 = x2, x1
            mask[x1:x2] = False
        self.xmasks.append(mask)
        self.plot_results()

    def on_undo(self, event=None):
        if len(self.xmasks) == 1:
            self.xmasks = [np.ones(len(self.data[0]), dtype=bool)]
        else:
            self.xmasks.pop()
        self.plot_results()

    def on_apply(self, event=None):
        xplot, yplot = self.get_xydata(datatype='xydata')
        mask = self.xmasks[-1]
        dgroup = self.dgroup
        energies_removed  = xplot[np.where(~mask)].tolist()
        dgroup.energy = dgroup.xplot = xplot[mask]
        dgroup.mu     = dgroup.yplot = yplot[mask]
        self.reset_data_history()
        dgroup.journal.add('deglitch_removed_energies', energies_removed)
        self.parent.process_normalization(dgroup)
        self.plot_results()

    def on_saveas(self, event=None):
        fname = self.wids['grouplist'].GetStringSelection()
        new_fname = self.wids['save_as_name'].GetValue()
        ngroup = self.controller.copy_group(fname, new_filename=new_fname)
        xplot, yplot = self.get_xydata(datatype='mu')
        mask = self.xmasks[-1]
        energies_removed  = xplot[np.where(~mask)].tolist()

        ngroup.energy = ngroup.xplot = xplot[mask]
        ngroup.mu     = ngroup.yplot = yplot[mask]
        ngroup.energy_orig = 1.0*ngroup.energy

        ogroup = self.controller.get_group(fname)
        olddesc = ogroup.journal.get('source_desc').value

        self.parent.install_group(ngroup, journal=ogroup.journal)
        ngroup.journal.add('source_desc', f"deglitched({olddesc})")
        ngroup.journal.add('deglitch_removed_energies', energies_removed)

        self.parent.process_normalization(ngroup)

    def plot_results(self, event=None, keep_limits=True):
        ppanel = self.controller.get_display(stacked=False).panel

        xplot, yplot = self.data

        xmin = min(xplot) - 0.025*(max(xplot) - min(xplot))
        xmax = max(xplot) + 0.025*(max(xplot) - min(xplot))
        ymin = min(yplot) - 0.025*(max(yplot) - min(yplot))
        ymax = max(yplot) + 0.025*(max(yplot) - min(yplot))

        dgroup = self.dgroup

        path, fname = path_split(dgroup.filename)

        plotstr = self.wids['plotopts'].GetStringSelection()
        plottype = DEGLITCH_PLOTS[plotstr]

        xlabel=plotlabels.energy
        if plottype in ('chie', 'chiew'):
            xmin = self.dgroup.e0
            xlabel = xlabel=plotlabels.ewithk

        opts = dict(xlabel=xlabel, title='De-glitching:\n %s' % fname,
                    delay_draw=True)

        ylabel =  {'mu': plotlabels.mu,
                   'norm': plotlabels.norm,
                   'chie':  plotlabels.chie,
                   'chiew': plotlabels.chiew.format(1),
                   }.get(plottype, plotlabels.norm)

        dgroup.plot_xlabel = xlabel
        dgroup.plot_ylabel = ylabel

        xlim, ylim = get_view_limits(ppanel)

        ppanel.plot(xplot, yplot, zorder=10, marker=None, linewidth=3,
                    label='original', ylabel=ylabel, **opts)

        if len(self.xmasks) > 1:
            mask = self.xmasks[-1]
            ppanel.oplot(xplot[mask], yplot[mask], zorder=15,
                         marker='o', markersize=3, linewidth=2.0,
                         label='current', show_legend=True, **opts)

        def ek_formatter(x, pos):
            ex = float(x) - self.dgroup.e0
            s = '' if ex < 0 else '\n[%.1f]' % (etok(ex))
            return r"%1.4g%s" % (x, s)

        if keep_limits:
            set_view_limits(ppanel, xlim, ylim)
        if plottype in ('chie', 'chiew'):
            ppanel.axes.xaxis.set_major_formatter(FuncFormatter(ek_formatter))

        if self.plot_markers is not None:
            rchoice = self.choice_range.GetStringSelection().lower()
            xwork, ywork = self.data
            opts = dict(marker='o', markersize=6, zorder=2, label='_nolegend_',
                        markerfacecolor='#66000022', markeredgecolor='#440000')
            if self.plot_markers == 'xlast':
                bad = index_nearest(xwork, self.wids['xlast'].GetValue())
                ppanel.axes.plot([xwork[bad]], [ywork[bad]], **opts)
            else:
                bad = index_nearest(xwork, self.wids['range1'].GetValue())
                if rchoice == 'above':
                    ppanel.axes.plot([xwork[bad:]], [ywork[bad:]], **opts)
                elif rchoice == 'below':
                    ppanel.axes.plot([xwork[:bad+1]], [ywork[:bad+1]], **opts)
                elif rchoice == 'between':
                    bad2 = index_nearest(xwork, self.wids['range2'].GetValue())
                    ppanel.axes.plot([xwork[bad:bad2+1]],
                                     [ywork[bad:bad2+1]], **opts)


        ppanel.canvas.draw()

        self.history_message.SetLabel('%i items in history' % (len(self.xmasks)-1))

    def GetResponse(self):
        raise AttributeError("use as non-modal dialog!")


SPECCALC_SETUP = """#From SpectraCalc dialog:
_x = {group:s}.{xname:s}
a = {group:s}.{yname:s}
b = c = d = e = f = g = None
"""

SPECCALC_INTERP = "{key:s} = interp({group:s}.{xname:s}, {group:s}.{yname:s}, _x)"
SPECCALC_PLOT = """plot(_x, ({expr:s}), label='{expr:s}', new=True,
   show_legend=True, xlabel='{xname:s}', title='Spectral Calculation')"""

SPECCALC_SAVE = """{new:s} = copy_xafs_group({group:s})
{new:s}.groupname = '{new:s}'
{new:s}.mu = ({expr:s})
{new:s}.filename = '{fname:s}'
{new:s}.journal.add('calc_groups', {group_map:s})
{new:s}.journal.add('calc_arrayname', '{yname:s}')
{new:s}.journal.add('calc_expression', '{expr:s}')
del _x, a, b, c, d, e, f, g"""


class SpectraCalcDialog(wx.Dialog):
    """dialog for adding and subtracting spectra"""
    def __init__(self, parent, controller, **kws):

        self.parent = parent
        self.controller = controller
        self.dgroup = self.controller.get_group()
        self.group_a = None
        groupnames = list(self.controller.file_groups.keys())

        self.data = [self.dgroup.energy[:], self.dgroup.norm[:]]
        xmin = min(self.dgroup.energy)
        xmax = max(self.dgroup.energy)
        e0val = getattr(self.dgroup, 'e0', xmin)

        wx.Dialog.__init__(self, parent, wx.ID_ANY, size=(475, 525),
                           title="Spectra Calculations: Add, Subtract Spectra")
        self.SetFont(Font(FONTSIZE))
        panel = GridPanel(self, ncols=3, nrows=4, pad=4, itemstyle=LEFT)

        def add_text(text, dcol=1, newrow=True):
            panel.Add(SimpleText(panel, text), dcol=dcol, newrow=newrow)

        self.wids = wids = {}
        array_choices = ('Normalized \u03BC(E)', 'Raw \u03BC(E)')

        wids['array'] = Choice(panel, choices=array_choices, size=(250, -1))

        add_text('Array to use: ',  newrow=True)
        panel.Add(wids['array'], dcol=2)

        # group 'a' cannot be none, and defaults to current group
        gname = 'a'
        wname = 'group_%s' % gname
        wids[wname] = Choice(panel, choices=groupnames, size=(250, -1))
        wids[wname].SetStringSelection(self.dgroup.filename)
        add_text('   %s = ' % gname,  newrow=True)
        panel.Add(wids[wname], dcol=2)

        groupnames.insert(0, 'None')
        for gname in ('b', 'c', 'd', 'e', 'f', 'g'):
            wname = 'group_%s' % gname
            wids[wname] = Choice(panel, choices=groupnames, size=(250, -1))
            wids[wname].SetSelection(0)
            add_text('   %s = ' % gname,  newrow=True)
            panel.Add(wids[wname], dcol=2)

        wids['formula'] = wx.TextCtrl(panel, -1, 'a-b', size=(250, -1))
        add_text('Expression = ',  newrow=True)
        panel.Add(wids['formula'], dcol=2)

        wids['docalc'] = Button(panel, 'Calculate',
                                size=(150, -1), action=self.on_docalc)

        panel.Add(wids['docalc'], dcol=2, newrow=True)

        wids['save_as'] = Button(panel, 'Save As New Group: ', size=(150, -1),
                           action=self.on_saveas)
        SetTip(wids['save_as'], 'Save as new group')

        wids['save_as_name'] = wx.TextCtrl(panel, -1,
                                           self.dgroup.filename + '_calc',
                                           size=(250, -1))
        panel.Add(wids['save_as'], newrow=True)
        panel.Add(wids['save_as_name'], dcol=2)
        wids['save_as'].Disable()
        panel.Add(Button(panel, 'Done', size=(150, -1), action=self.onDone),
                  newrow=True)
        panel.pack()
        fit_dialog_window(self, panel)

    def onDone(self, event=None):
        self.Destroy()

    def on_docalc(self, event=None):
        self.expr = self.wids['formula'].GetValue()

        self.yname = 'mu'
        if self.wids['array'].GetStringSelection().lower().startswith('norm'):
            self.yname = 'norm'

        groups = {}
        for aname in ('a', 'b', 'c', 'd', 'e', 'f', 'g'):
            fname = self.wids['group_%s' % aname].GetStringSelection()
            if fname not in (None, 'None'):
                grp = self.controller.get_group(fname)
                groups[aname] = grp

        self.group_map = {key: group.groupname for key, group in groups.items()}
        # note: 'a' cannot be None, all others can be None
        group_a = self.group_a = groups.pop('a')
        xname = 'energy'
        if not hasattr(group_a, xname):
            xname = 'xplot'

        cmds = [SPECCALC_SETUP.format(group=group_a.groupname,
                                      xname=xname, yname=self.yname)]

        for key, group in groups.items():
            cmds.append(SPECCALC_INTERP.format(key=key, group=group.groupname,
                                               xname=xname, yname=self.yname))

        cmds.append(SPECCALC_PLOT.format(expr=self.expr, xname=xname))
        self.controller.larch.eval('\n'.join(cmds))
        self.wids['save_as'].Enable()

    def on_saveas(self, event=None):
        wids = self.wids
        _larch = self.controller.larch
        fname = wids['group_a'].GetStringSelection()
        new_fname =self.wids['save_as_name'].GetValue()
        new_gname = file2groupname(new_fname, slen=5, symtable=_larch.symtable)

        gmap = []
        for k, v in self.group_map.items():
            gmap.append(f'"{k}": "{v}"')
        gmap = '{%s}' % (', '.join(gmap))

        _larch.eval(SPECCALC_SAVE.format(new=new_gname, fname=new_fname,
                                         group=self.group_a.groupname,
                                         group_map=gmap,
                                         yname=self.yname, expr=self.expr))


        journal={'source_desc': f"{new_fname}: calc({self.expr})",
                 'calc_groups': gmap,  'calc_expression': self.expr}

        ngroup = getattr(_larch.symtable, new_gname, None)
        if ngroup is not None:
            self.parent.install_group(ngroup, source=journal['source_desc'],
                                      journal=journal)

    def GetResponse(self):
        raise AttributeError("use as non-modal dialog!")

class EnergyUnitsDialog(wx.Dialog):
    """dialog for selecting, changing energy units, forcing data to eV"""
    unit_choices = ['eV', 'keV', 'deg', 'steps']

    def __init__(self, parent, energy_array, unitname='eV',dspace=1, **kws):

        self.parent = parent
        self.energy = 1.0*energy_array

        title = "Select Energy Units to convert to 'eV'"
        wx.Dialog.__init__(self, parent, wx.ID_ANY, title=title)
        self.SetFont(Font(FONTSIZE))
        panel = GridPanel(self, ncols=3, nrows=4, pad=2, itemstyle=LEFT)

        self.en_units = Choice(panel, choices=self.unit_choices, size=(125, -1),
                               action=self.onUnits)
        self.en_units.SetStringSelection(unitname)
        self.mono_dspace = FloatCtrl(panel, value=dspace, minval=0, maxval=100.0,
                                     precision=6, size=(125, -1))
        self.steps2deg  = FloatCtrl(panel, value=1.0, minval=0,
                                     precision=1, size=(125, -1))

        self.mono_dspace.Disable()
        self.steps2deg.Disable()

        panel.Add(SimpleText(panel, 'Energy Units : '), newrow=True)
        panel.Add(self.en_units)

        panel.Add(SimpleText(panel, 'Mono D spacing : '), newrow=True)
        panel.Add(self.mono_dspace)

        panel.Add(SimpleText(panel, 'Mono Steps per Degree : '), newrow=True)
        panel.Add(self.steps2deg)
        panel.Add((5, 5))

        panel.Add(OkCancel(panel), dcol=2, newrow=True)
        panel.pack()

        fit_dialog_window(self, panel)


    def onUnits(self, event=None):
        units = self.en_units.GetStringSelection()
        self.steps2deg.Enable(units == 'steps')
        self.mono_dspace.Enable(units in ('steps', 'deg'))

    def GetResponse(self, master=None, gname=None, ynorm=True):
        self.Raise()
        response = namedtuple('EnergyUnitsResponse',
                              ('ok', 'units', 'energy', 'dspace'))
        ok, units, en, dspace = False, 'eV', None, -1

        if self.ShowModal() == wx.ID_OK:
            units = self.en_units.GetStringSelection()
            if units == 'eV':
                en = self.energy
            elif units == 'keV':
                en = self.energy * 1000.0
            elif units in ('steps', 'deg'):
                dspace = float(self.mono_dspace.GetValue())
                if units == 'steps':
                    self.energy /= self.steps2deg.GetValue()
                en = PLANCK_HC/(2*dspace*np.sin(self.energy * DEG2RAD))
            ok = True
        return response(ok, units, en, dspace)

class MergeDialog(wx.Dialog):
    """dialog for merging groups"""
    ychoices = ['raw mu(E)', 'normalized mu(E)']

    def __init__(self, parent, groupnames, outgroup='merge', **kws):
        title = "Merge %i Selected Groups" % (len(groupnames))
        wx.Dialog.__init__(self, parent, wx.ID_ANY, title=title)

        panel = GridPanel(self, ncols=3, nrows=4, pad=2, itemstyle=LEFT)

        self.master_group = Choice(panel, choices=groupnames, size=(250, -1))
        self.yarray_name  = Choice(panel, choices=self.ychoices, size=(250, -1))
        self.group_name   = wx.TextCtrl(panel, -1, outgroup,  size=(250, -1))

        panel.Add(SimpleText(panel, 'Match Energy to : '), newrow=True)
        panel.Add(self.master_group)

        panel.Add(SimpleText(panel, 'Array to merge  : '), newrow=True)
        panel.Add(self.yarray_name)

        panel.Add(SimpleText(panel, 'New group name  : '), newrow=True)
        panel.Add(self.group_name)

        panel.Add(OkCancel(panel), dcol=2, newrow=True)

        panel.pack()
        fit_dialog_window(self, panel)


    def GetResponse(self, master=None, gname=None, ynorm=True):
        self.Raise()
        response = namedtuple('MergeResponse', ('ok', 'master', 'ynorm', 'group'))
        ok = False
        if self.ShowModal() == wx.ID_OK:
            master= self.master_group.GetStringSelection()
            ynorm = 'norm' in self.yarray_name.GetStringSelection().lower()
            gname = self.group_name.GetValue()
            ok = True
        return response(ok, master, ynorm, gname)


class ExportCSVDialog(wx.Dialog):
    """dialog for exporting groups to CSV file"""

    def __init__(self, parent, groupnames, **kws):
        title = "Export Selected Groups"
        wx.Dialog.__init__(self, parent, wx.ID_ANY, title=title)
        self.SetFont(Font(FONTSIZE))
        self.xchoices = {'Energy': 'energy',
                         'k': 'k',
                         'R': 'r',
                         'q': 'q'}

        self.ychoices = {'normalized mu(E)': 'norm',
                         'raw mu(E)': 'mu',
                         'flattened mu(E)': 'flat',
                         'd mu(E) / dE': 'dmude',
                         'chi(k)': 'chi',
                         'chi(E)': 'chie',
                         'chi(q)': 'chiq',
                         '|chi(R)|': 'chir_mag',
                         'Re[chi(R)]': 'chir_re'}

        self.delchoices = {'comma': ',',
                           'space': ' ',
                           'tab': '\t'}

        panel = GridPanel(self, ncols=3, nrows=4, pad=2, itemstyle=LEFT)
        self.save_individual_files = Check(panel, default=False, label='Save individual files', action=self.onSaveIndividualFiles)
        self.master_group = Choice(panel, choices=groupnames, size=(200, -1))
        self.xarray_name  = Choice(panel, choices=list(self.xchoices.keys()), size=(200, -1))
        self.yarray_name  = Choice(panel, choices=list(self.ychoices.keys()), action=self.onYChoice, size=(200, -1))
        self.del_name     = Choice(panel, choices=list(self.delchoices.keys()), size=(200, -1))

        panel.Add(self.save_individual_files, newrow=True)

        panel.Add(SimpleText(panel, 'Group for Energy Array: '), newrow=True)
        panel.Add(self.master_group)

        panel.Add(SimpleText(panel, 'X Array to Export: '), newrow=True)
        panel.Add(self.xarray_name)

        panel.Add(SimpleText(panel, 'Y Array to Export: '), newrow=True)
        panel.Add(self.yarray_name)

        panel.Add(SimpleText(panel, 'Delimeter for File: '), newrow=True)
        panel.Add(self.del_name)
        panel.Add(OkCancel(panel), dcol=2, newrow=True)
        panel.pack()
        fit_dialog_window(self, panel)

    def onYChoice(self, event=None):
        ychoice = self.yarray_name.GetStringSelection()
        yval = self.ychoices[ychoice]
        xarray = 'Energy'
        if yval in ('chi', 'chiq'):
            xarray = 'k'
        elif yval in ('chir_mag', 'chir_re'):
            xarray = 'R'
        self.xarray_name.SetStringSelection(xarray)

    def onSaveIndividualFiles(self, event=None):
        save_individual = self.save_individual_files.IsChecked()
        self.master_group.Enable(not save_individual)

    def GetResponse(self, master=None, gname=None, ynorm=True):
        self.Raise()
        response = namedtuple('ExportCSVResponse',
                              ('ok', 'individual', 'master', 'xarray', 'yarray', 'delim'))
        ok = False
        individual = master = ''
        xarray, yarray, delim = 'Energy', '', ','
        if self.ShowModal() == wx.ID_OK:
            individual = self.save_individual_files.IsChecked()
            master = self.master_group.GetStringSelection()
            xarray = self.xchoices[self.xarray_name.GetStringSelection()]
            yarray = self.ychoices[self.yarray_name.GetStringSelection()]
            delim  = self.delchoices[self.del_name.GetStringSelection()]
            ok = True
        return response(ok, individual, master, xarray, yarray, delim)

class QuitDialog(wx.Dialog):
    """dialog for quitting, prompting to save project"""

    def __init__(self, parent, message, **kws):
        title = "Quit Larch Larix?"
        wx.Dialog.__init__(self, parent, wx.ID_ANY, title=title, size=(500, 150))
        self.SetFont(Font(FONTSIZE))
        self.needs_save = True
        panel = GridPanel(self, ncols=3, nrows=4, pad=2, itemstyle=LEFT)

        status, filename, stime = message
        warn_msg = 'All work in this session will be lost!'

        panel.Add((5, 5))
        if len(stime) > 2:
            status = f"{status} at {stime} to file"
            warn_msg = 'Changes made after that will be lost!'

        panel.Add(wx.StaticText(panel, label=status), dcol=2)

        if len(filename) > 0:
            if filename.startswith("'") and filename.endswith("'"):
                filename = filename[1:-1]
            panel.Add((15, 5), newrow=True)
            panel.Add(wx.StaticText(panel, label=filename), dcol=2)

        panel.Add((5, 5), newrow=True)
        panel.Add(wx.StaticText(panel, label=warn_msg), dcol=2)
        panel.Add(HLine(panel, size=(500, 3)), dcol=3, newrow=True)
        panel.Add((5, 5), newrow=True)
        panel.Add(OkCancel(panel), dcol=2, newrow=True)
        panel.pack()

        fit_dialog_window(self, panel)

    def GetResponse(self):
        self.Raise()
        response = namedtuple('QuitResponse', ('ok',))
        ok = (self.ShowModal() == wx.ID_OK)
        return response(ok,)

class RenameDialog(wx.Dialog):
    """dialog for renaming group"""
    def __init__(self, parent, oldname,  **kws):
        title = "Rename Group %s" % (oldname)
        wx.Dialog.__init__(self, parent, wx.ID_ANY, title=title)
        self.SetFont(Font(FONTSIZE))
        panel = GridPanel(self, ncols=3, nrows=4, pad=2, itemstyle=LEFT)

        self.newname   = wx.TextCtrl(panel, -1, oldname,  size=(250, -1))

        panel.Add(SimpleText(panel, 'Old Name : '), newrow=True)
        panel.Add(SimpleText(panel, oldname))
        panel.Add(SimpleText(panel, 'New Name : '), newrow=True)
        panel.Add(self.newname)
        panel.Add(OkCancel(panel), dcol=2, newrow=True)

        panel.pack()
        fit_dialog_window(self, panel)


    def GetResponse(self, newname=None):
        self.Raise()
        response = namedtuple('RenameResponse', ('ok', 'newname'))
        ok = False
        if self.ShowModal() == wx.ID_OK:
            newname = self.newname.GetValue()
            ok = True
        return response(ok, newname)

class RemoveDialog(wx.Dialog):
    """dialog for removing groups"""
    def __init__(self, parent, grouplist,  **kws):
        title = "Remove %i Selected Group" % len(grouplist)
        self.grouplist = grouplist
        wx.Dialog.__init__(self, parent, wx.ID_ANY, title=title)
        self.SetFont(Font(FONTSIZE))
        panel = GridPanel(self, ncols=3, nrows=4, pad=2, itemstyle=LEFT)

        panel.Add(SimpleText(panel, 'Remove %i Selected Groups?' % (len(grouplist))),
                  newrow=True, dcol=2)

        panel.Add(OkCancel(panel), dcol=2, newrow=True)
        panel.pack()
        fit_dialog_window(self, panel)

    def GetResponse(self, ngroups=None):
        self.Raise()
        response = namedtuple('RemoveResponse', ('ok','ngroups'))
        ok = False
        if self.ShowModal() == wx.ID_OK:
            ngroups = len(self.grouplist)
            ok = True
        return response(ok, ngroups)


class LoadSessionDialog(wx.Frame):
    """Read, show data from saved larch session"""

    xasgroups_name = '_xasgroups'
    feffgroups_name = ['_feffpaths', '_feffcache']

    def __init__(self, parent, session, filename, controller, **kws):
        self.parent = parent
        self.session = session
        self.filename = filename
        self.controller = controller
        title = f"Read Larch Session from '{filename}'"
        wx.Frame.__init__(self, parent, wx.ID_ANY, title=title)

        x0, y0 = parent.GetPosition()
        self.SetPosition((x0+450, y0+75))

        splitter  = wx.SplitterWindow(self, style=wx.SP_LIVE_UPDATE)
        splitter.SetMinimumPaneSize(250)

        leftpanel = wx.Panel(splitter)
        rightpanel = wx.Panel(splitter)

        ltop = wx.Panel(leftpanel)

        sel_none = Button(ltop, 'Select None', size=(100, 30), action=self.onSelNone)
        sel_all  = Button(ltop, 'Select All', size=(100, 30), action=self.onSelAll)
        sel_imp  = Button(ltop, 'Import Selected Data', size=(200, 30),
                          action=self.onImport)

        self.select_imported = sel_imp
        self.grouplist = FileCheckList(leftpanel, select_action=self.onShowGroup)
        set_color(self.grouplist, 'list_fg', bg='list_bg')

        tsizer = wx.GridBagSizer(2, 2)
        tsizer.Add(sel_all, (0, 0), (1, 1), LEFT, 0)
        tsizer.Add(sel_none,  (0, 1), (1, 1), LEFT, 0)
        tsizer.Add(sel_imp,  (1, 0), (1, 2), LEFT, 0)

        pack(ltop, tsizer)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(ltop, 0, LEFT|wx.GROW, 1)
        sizer.Add(self.grouplist, 1, LEFT|wx.GROW|wx.ALL, 1)
        pack(leftpanel, sizer)


        panel = GridPanel(rightpanel, ncols=3, nrows=4, pad=2, itemstyle=LEFT)
        self.wids = wids = {}

        top_message = 'Larch Session File: No XAFS Groups'
        symtable = controller.symtable

        self.allgroups = session.symbols.get(self.xasgroups_name, {})
        self.extra_groups = []
        for key, val in session.symbols.items():
            if key == self.xasgroups_name or key in self.feffgroups_name:
                continue
            if key in self.allgroups:
                continue
            if hasattr(val, 'energy') and hasattr(val, 'mu'):
                if key in self.allgroups.keys() or key in self.allgroups.values():
                    continue
                self.allgroups[key] = key
                self.extra_groups.append(key)


        checked = []
        for fname, gname in self.allgroups.items():
            self.grouplist.Append(fname)
            checked.append(fname)

        self.grouplist.SetCheckedStrings(checked)

        group_names = list(self.allgroups.values())
        group_names.append(self.xasgroups_name)
        group_names.extend(self.feffgroups_name)

        wids['view_conf'] = Button(panel, 'Show Session Configuration',
                                     size=(200, 30), action=self.onShowConfig)
        wids['view_cmds'] = Button(panel, 'Show Session Commands',
                                     size=(200, 30), action=self.onShowCommands)

        wids['plotopt'] = Choice(panel, choices=list(SESSION_PLOTS.keys()),
                                 action=self.onPlotChoice, size=(175, -1))

        panel.Add(wids['view_conf'], dcol=1)
        panel.Add(wids['view_cmds'], dcol=1, newrow=False)
        panel.Add(HLine(panel, size=(450, 2)), dcol=3, newrow=True)

        over_msg = 'Importing these Groups/Data will overwrite values in the current session:'
        panel.Add(SimpleText(panel, over_msg), dcol=2, newrow=True)
        panel.Add(SimpleText(panel, "Symbol Name"), dcol=1, newrow=True)
        panel.Add(SimpleText(panel, "Import/Overwrite?"), dcol=1)
        i = 0
        self.overwrite_checkboxes = {}
        for g in self.session.symbols:
            if g not in group_names and hasattr(symtable, g):
                chbox = Check(panel, default=True)
                panel.Add(SimpleText(panel, g),  dcol=1, newrow=True)
                panel.Add(chbox,  dcol=1)
                self.overwrite_checkboxes[g] = chbox

                i += 1

        panel.Add((5, 5), newrow=True)
        panel.Add(HLine(panel, size=(450, 2)), dcol=3, newrow=True)
        panel.Add(SimpleText(panel, 'Plot Type:'), newrow=True)
        panel.Add(wids['plotopt'], dcol=2, newrow=False)
        panel.pack()

        self.plotpanel = PlotPanel(rightpanel, messenger=self.plot_messages)
        self.plotpanel.SetSize((475, 450))
        plotconf = self.controller.get_config('plot')
        self.plotpanel.conf.set_theme(plotconf['theme'])
        self.plotpanel.conf.enable_grid(plotconf['show_grid'])

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(panel, 0, LEFT, 2)
        sizer.Add(self.plotpanel, 1, LEFT, 2)

        pack(rightpanel, sizer)

        splitter.SplitVertically(leftpanel, rightpanel, 1)
        self.SetSize((750, 725))

        self.Show()
        self.Raise()

    def plot_messages(self, msg, panel=1):
        pass

    def onSelAll(self, event=None):
        self.grouplist.SetCheckedStrings(list(self.allgroups.keys()))

    def onSelNone(self, event=None):
        self.grouplist.SetCheckedStrings([])

    def onShowGroup(self, event=None):
        """column selections changed calc xplot and yplot"""
        fname = event.GetString()
        gname = self.allgroups.get(fname, None)
        if gname in self.session.symbols:
            self.plot_group(gname, fname)

    def onPlotChoice(self, event=None):
        fname = self.grouplist.GetStringSelection()
        gname = self.allgroups.get(fname, None)
        self.plot_group(gname, fname)

    def plot_group(self, gname, fname):
        grp = self.session.symbols[gname]
        plottype = SESSION_PLOTS.get(self.wids['plotopt'].GetStringSelection(), 'norm')
        xdef = np.zeros(1)
        xplot = getattr(grp, 'energy', xdef)
        yplot = getattr(grp, 'mu', xdef)
        xlabel = plotlabels.energy
        ylabel = plotlabels.mu
        if plottype == 'norm' and hasattr(grp, 'norm'):
            yplot = getattr(grp, 'norm', xdef)
            ylabel = plotlabels.norm
        elif plottype == 'chikw' and hasattr(grp, 'chi'):
            xplot = getattr(grp, 'k', xdef)
            yplot = getattr(grp, 'chi', xdef)
            yplot = yplot*xplot*xplot
            xlabel = plotlabels.chikw.format(2)

        if len(yplot) > 1:
            self.plotpanel.plot(xplot, yplot, xlabel=xlabel,
                                ylabel=ylabel, title=fname)


    def onShowConfig(self, event=None):
        DictFrame(parent=self.parent,
                  data=self.session.config,
                  title=f"Session Configuration for '{self.filename}'")

    def onShowCommands(self, event=None):
        oname = self.filename.replace('.larix', '.lar')
        wildcard='Larch Command Files (*.lar)|*.lar'
        text = '\n'.join(self.session.command_history)
        ReportFrame(parent=self.parent,
                    text=text,
                    title=f"Session Commands from '{self.filename}'",
                    default_filename=oname,
                    wildcard=wildcard)

    def onClose(self, event=None):
        self.Destroy()

    def onImport(self, event=None):
        ignore = []
        for gname, chbox in self.overwrite_checkboxes.items():
            if not chbox.IsChecked():
                ignore.append(gname)

        sel_groups = self.grouplist.GetCheckedStrings()
        for fname, gname in self.allgroups.items():
            if fname not in sel_groups:
                ignore.append(gname)

        fname = Path(self.filename).as_posix()
        if fname.endswith('/'):
            fname = fname[:-1]
        lcmd = [f"load_session('{fname}'"]
        if len(ignore) > 0:
            ignore = repr(ignore)
            lcmd.append(f", ignore_groups={ignore}")
        if len(self.extra_groups) > 0:
            extra = repr(self.extra_groups)
            lcmd.append(f", include_xasgroups={extra}")

        lcmd = ''.join(lcmd) + ')'

        cmds = ["# Loading Larch Session with ", lcmd, '######']

        self.controller.larch.eval('\n'.join(cmds))
        last_fname = None
        xasgroups = getattr(self.controller.symtable, self.xasgroups_name, {})
        for key, val in xasgroups.items():
            if key not in self.controller.filelist.GetItems():
                self.controller.filelist.Append(key)
                last_fname = key

        self.controller.recentfiles.append((time.time(), self.filename))

        wx.CallAfter(self.Destroy)
        if last_fname is not None:
            self.parent.ShowFile(filename=last_fname)



import codecs
from os import path

from docutils.io import StringOutput

from sphinx.util.console import bold, darkgreen, brown
from sphinx.util.nodes import inline_all_toctrees
from sphinx.util.osutil import ensuredir, os_path
from sphinx.builders.text import TextBuilder
from sphinx.writers.text import TextWriter, TextTranslator

class SingleFileTextTranslator(TextTranslator):
    def visit_start_of_file(self, node):
        pass
    def depart_start_of_file(self, node):
        pass

class SingleFileTextWriter(TextWriter):
    def translate(self):
        visitor = SingleFileTextTranslator(self.document, self.builder)
        self.document.walkabout(visitor)
        self.output = visitor.body

class SingleFileTextBuilder(TextBuilder):
    """
    A TextBuilder subclass that puts the whole document tree in one text file.
    """
    name = 'singletext'
    copysource = False

    def get_outdated_docs(self):
        return 'all documents'

    def assemble_doctree(self):
        master = self.config.master_doc
        tree = self.env.get_doctree(master)
        tree = inline_all_toctrees(
            self, set(), master, tree, darkgreen, [master])
        tree['docname'] = master
        self.env.resolve_references(tree, master, self)
        return tree

    def prepare_writing(self, docnames):
        self.writer = SingleFileTextWriter(self)

    def write(self, *ignored):
        docnames = self.env.all_docs

        self.info(bold('preparing documents... '), nonl=True)
        self.prepare_writing(docnames)
        self.info('done')

        self.info(bold('assembling single document... '), nonl=True)
        doctree = self.assemble_doctree()
        self.info()
        self.info(bold('writing... '), nonl=True)
        self.write_doc(self.config.master_doc, doctree)
        self.info('done')

def setup(app):
    app.add_builder(SingleFileTextBuilder)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Utilities to work with 2D grids and interpolation
=================================================
"""
from __future__ import division, print_function

import warnings
import numpy as np

# suppress warnings
try:
    import numexpr
except ImportError:
    pass

from larch.utils.logging import getLogger

_logger = getLogger("gridxyz")

### GLOBAL VARIABLES ###
MODNAME = "_math"


def gridxyz(xcol, ycol, zcol, xystep=None, lib="scipy", method="linear"):
    """Grid (X, Y, Z) 1D data on a 2D regular mesh

    Parameters
    ----------
    xcol, ycol, zcol : numpy.ndarray
        1D arrays representing the map (z is the signal intensity)

    xystep : float or None (optional)
        the step size of the XY grid, if None -> 0.1

    lib : str, optional ["scipy"]
        library used for griddata
            - scipy
            - matplotlib

    method : str, optional ["cubic"]
        interpolation method
            - nearest
            - linear
            - cubic

    Returns
    -------
        xgrid, ygrid, zz : numpy.ndarray
            xgrid, ygrid : 1D arrays giving abscissa and ordinate of the map
            zz : 2D array with the gridded intensity map

    See also
    --------
        - MultipleScanToMeshPlugin in PyMca
    """
    if xystep is None:
        xystep = 0.1
        _logger.warning(
            "'xystep' not given: using a default value of {0}".format(xystep)
        )
    assert type(xystep) is float, "xystep should be float"
    # create the XY meshgrid and interpolate the Z on the grid
    nxpoints = int((xcol.max() - xcol.min()) / xystep)
    nypoints = int((ycol.max() - ycol.min()) / xystep)
    xgrid = np.linspace(xcol.min(), xcol.max(), num=nxpoints)
    ygrid = np.linspace(ycol.min(), ycol.max(), num=nypoints)
    xx, yy = np.meshgrid(xgrid, ygrid)
    if "matplotlib" in lib.lower():
        _logger.warning("matplotlib deprecated -> using scipy")
        lib = "scipy"
    if not "scipy" in lib.lower():
        _logger.error("lib should be scipy")
        return np.nan, np.nan, np.nan

    try:
        from scipy.interpolate import griddata
    except ImportError:
        _logger.error("Cannot load griddata from Scipy")
        return np.nan, np.nan, np.nan

    _logger.info("Gridding data with {0}/{1}...".format(lib, method))
    zz = griddata(
        (xcol, ycol),
        zcol,
        (xgrid[None, :], ygrid[:, None]),
        method=method,
        fill_value=np.nan,
    )


    return xgrid, ygrid, zz


if __name__ == "__main__":
    pass



#!/usr/bin/env python
"""
site configuration for larch:

   init_files:  list of larch files run (in order) on startup
   history_file:
"""
import sys
import os
from pathlib import Path

from subprocess import check_call, CalledProcessError, TimeoutExpired

from packaging.version import parse as version_parse

from .utils import (uname, get_homedir, log_warning, log_error)
from .version import __version__, __release_version__

larch_version = __version__
larch_release_version = __release_version__

# lists of recommended packages that are not installed by default
# but may be installed if several of the larch apps are run.

def pjoin(*args):
    "simple join"
    return Path(*args).absolute().as_posix()

def update_larch(with_larix=True):
    "pip upgrade larch"
    target = 'xraylarch'
    if with_larix:
        target = 'xraylarch[larix]'
    check_call([sys.executable, '-m', 'pip', 'install', '--upgrade', target])

#                 = get_homedir() + 'larch'  (#win)
home_dir = get_homedir()

icondir = Path(Path(__file__).parent, 'icons').absolute()

user_larchdir = pjoin(home_dir, '.larch')

if 'LARCHDIR' in os.environ:
    user_larchdir = Path(os.environ['LARCHDIR']).absolute().as_posix()

# on Linux, check for HOME/.local/share,
# make with mode=711 if needed
if uname in ('linux', 'darwin') and os.getuid() > 0:
    lshare = Path(home_dir, '.local', 'share').absolute()
    lshare.mkdir(mode=457, parents=True, exist_ok=True) # for octal 711


# initialization larch files to be run on startup
init_files = [pjoin(user_larchdir, 'init.lar')]

if 'LARCHSTARTUP' in os.environ:
    startup = Path(os.environ['LARCHSTARTUP'])
    if startup.exists():
        init_files = [startup.as_posix()]

# history file:
history_file = pjoin(user_larchdir, 'history.lar')

def make_user_larchdirs():
    """create user's larch directories"""
    files = {'init.lar':             'put custom startup larch commands:',
             'history.lar':          'history of commands for larch CLI',
             'history_larchgui.lar': 'history of commands for larch GUI',
             }
    subdirs = {'matplotlib': 'matplotlib may put cache files here',
               'feff':       'Feff files and subfolders here',
               'fdmnes':     'FDMNES files and subfolders here',
               }

    def make_dir(dname):
        "create directory"
        dname = Path(dname).absolute()
        if not dname.exists():
            try:
                dname.mkdir(mode=493, parents=True)
            except PermissionError:
                log_warning(f'no permission to create directory {dname.as_posix()}')
            except (OSError, TypeError):
                log_error(sys.exc_info()[1])

    def write_file(fname, text):
        "write wrapper"
        if not Path(fname).exists():
            try:
                with open(fname, 'w', encoding=sys.getdefaultencoding()) as fileh:
                    fileh.write(f'# {text}\n')
            except IOError:
                log_error(sys.exc_info()[1])

    make_dir(user_larchdir)
    for fname, text in files.items():
        write_file(pjoin(user_larchdir, fname), text)

    for sdir, text in subdirs.items():
        sdir = pjoin(user_larchdir, sdir)
        make_dir(sdir)
        write_file(pjoin(sdir, 'README'), text)


def site_config():
    "retutn string of site config"
    return f"""#== Larch Configuration:
  Release version:     {__release_version__}
  Development version: {__version__}
  Python executable:   {sys.executable}
  User larch dir:      {user_larchdir}
  User history_file:   {history_file}
#========================
"""

def show_site_config():
    "print stie_config"
    print(site_config())

def system_settings():
    """set system-specific Environmental Variables, and make sure
    that the user larchdirs exist.
    This is run by the interpreter on startup."""
    # ubuntu / unity hack
    if uname.startswith('linux'):
        if 'ubuntu' in os.uname()[3].lower():
            os.environ['UBUNTU_MENUPROXY'] = '0'
    make_user_larchdirs()


if __name__ == '__main__':
    show_site_config()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Struct2XAS: convert CIFs and XYZs files to FDMNES and FEFF inputs
"""
# main imports
import os
import json
import time
import tempfile
import numpy as np

# pymatgen
from pymatgen.core import Structure, Element, Lattice
from pymatgen.io.xyz import XYZ
from pymatgen.io.cif import CifParser
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

import larch.utils.logging as logging
from larch.utils import mkdir, unixpath
from larch.utils.strutils import fix_filename, unique_name, strict_ascii
from larch.site_config import user_larchdir
from larch.io import read_ascii
from larch.math.convolution1D import lin_gamma, conv

try:
    import pandas as pd
    from pandas.io.formats.style import Styler

    HAS_PANDAS = True
except ImportError:
    HAS_PANDAS = False

try:
    import py3Dmol

    HAS_PY3DMOL = True
except ImportError:
    HAS_PY3DMOL = False


__author__ = ["Beatriz G. Foschiani", "Mauro Rovezzi"]
__email__ = ["beatrizgfoschiani@gmail.com", "mauro.rovezzi@esrf.fr"]
__credits__ = ["Jade Chongsathapornpong", "Marius Retegan"]
__version__ = "2023.3.0"


# initialize the logger
logger = logging.getLogger("struct2xas", level="INFO")


def _get_timestamp() -> str:
    """return a custom time stamp string: YYY-MM-DD_HHMM"""
    return "{0:04d}-{1:02d}-{2:02d}_{3:02d}{4:02d}".format(*time.localtime())


def _pprint(matrix):
    """pretty print a list of lists (in case Pandas is not available)
    from: https://stackoverflow.com/questions/13214809/pretty-print-2d-list
    an alternative could be: https://pypi.org/project/tabulate/
    """
    s = [[str(e) for e in row] for row in matrix]
    lens = [max(map(len, col)) for col in zip(*s)]
    fmt = "\t".join("{{:{}}}".format(x) for x in lens)
    table = [fmt.format(*row) for row in s]
    print("\n".join(table))


def xyz2struct(molecule):
    """Convert pymatgen molecule to dummy pymatgen structure"""

    alat, blat, clat = 1, 1, 1

    # Set the lattice dimensions in each direction
    for i in range(len(molecule) - 1):
        if molecule.cart_coords[i][0] > molecule.cart_coords[i + 1][0]:
            alat = molecule.cart_coords[i][0]
        if molecule.cart_coords[i][1] > molecule.cart_coords[i + 1][1]:
            blat = molecule.cart_coords[i][1]
        if molecule.cart_coords[i][2] > molecule.cart_coords[i + 1][2]:
            clat = molecule.cart_coords[i][2]

    # Set the lattice dimensions in each direction
    lattice = Lattice.from_parameters(
        a=alat, b=blat, c=clat, alpha=90, beta=90, gamma=90
    )

    # Create a list of species
    species = [Element(sym) for sym in molecule.species]

    # Create a list of coordinates
    coords = molecule.cart_coords

    # Create the Structure object
    struct = Structure(lattice, species, coords, coords_are_cartesian=True)
    return struct


def structure_folders():
    """
    return folders under USER_LARCHDIR for FDMNES, Feff,  and structure files,
    making sure each exists
    """
    folders = {}
    for name in ("feff", "fdmnes", "mp_structs"):
        folders[name] = unixpath(os.path.join(user_larchdir, name))
        mkdir(folders[name])
    return folders


class Struct2XAS:
    """Class to convert data from CIF and XYZ files to FDMNES and FEFF inputs"""

    def __init__(self, file, abs_atom) -> None:
        """

        Arguments
        ---------
        file : str
            full path string to cif or xyz file
        abs_atom : str
            Absorber element in the structure, e.g.: "Fe", "Ni".

        Returns
        -------
        None

        ..note::

        --> IMPORTANT: <--

        for xyz files:
            Structures from xyz files are always considered non-symmetric for
            the lack of lattice information. For creating the object from an
            XYZ file, a non-symmetric `structure` object from pymatgen is
            generated (spacegroup: P1) from a `molecule` one via
            :func:`xyz2struct`. The lattice parameters chosen for this
            structure are arbitrary and are based on the size of the molecule,
            as are the fractional coordinates. Therefore, the analysis of this
            structure is limited to the central atoms and is not valid for
            atoms at the border of the molecule.

        for cif files:
            For creating the object from cif file, a pymatgen structure is
            generated with symmetry information from cif file.
        """

        self.file = file
        self.abs_atom = abs_atom
        self.frame = 0
        self.abs_site = 0
        self.is_cif = False
        self.is_xyz = False
        self.read_structure(file)
        self.nabs_sites = len(self.get_abs_sites())
        self.elems = self._get_elems()
        self.species = self._get_species()
        self.parent_path = user_larchdir
        self.folders = structure_folders()
        logger.info(
            f"Frames: {self.nframes}, Absorbing sites: {self.nabs_sites}. (Indexes for frames and abs_sites start at 0)"
        )

    def set_frame(self, frame=0):
        """For multiframe xyz file, set the frame index site."""
        self.frame = frame
        if self.molecules is not None:
            self.struct = xyz2struct(self.molecules[self.frame])
            self.mol = self.molecules[self.frame]
        else:
            logger.error("single frame structure")
        logger.debug(f"frame idx: {frame} ")

    def get_frame(self):
        """For multiframe xyz file, get the frame index site."""
        return self.frame

    def set_abs_site(self, abs_site=0):
        """Set the crystallographic index site for the absorbing atom."""
        try:
            self.get_abs_sites()[abs_site]
        except IndexError:
            logger.error("absorber site index out of range")
            raise IndexError("absorber site index out of range")
        logger.debug(f"abs_site idx: {abs_site} ")
        self.abs_site = abs_site

    def get_abs_site(self):
        """Get the crystallographic index site for the absorbing atom."""
        return self.abs_site

    def _get_elems(self):
        """Get elements present in the structure"""
        elems_list = []
        for _, elem in enumerate(self.struct):
            # if self.is_cif:
            for e in elem.species.elements:
                if e.name not in elems_list:
                    elems_list.append(e.name)

            # if self.is_xyz:
            #     if elem.species_string not in elems_list:
            #         elems_list.append(elem.species_string)

        return elems_list

    def _get_species(self):
        """Get elements present in the structure"""
        species_list = []
        species_list_occu = []
        for _, elem in enumerate(self.struct):
            if not self.full_occupancy:
                if str(elem.species) not in species_list_occu:
                    species_list_occu.append(str(elem.species))
            self.atoms_occu = species_list_occu

            for e in elem.species.elements:
                if str(e) not in species_list:
                    species_list.append(str(e))

        return species_list

    def read_structure(self, file):
        """Read and initialize the structure/molecule from the input file"""
        # Split the file name and extension
        self.file = file
        if os.path.isfile(self.file):
            # file_dirname = os.path.dirname(file)
            file = os.path.basename(self.file)
            split_file = os.path.splitext(file)
            self.file_name = split_file[0]
            self.file_ext = split_file[1]
        else:
            self.file_name, self.file_ext = None, None
            errmsg = f"{self.file} not found"
            logger.error(errmsg)
            raise FileNotFoundError(errmsg)

        if self.file_ext == ".cif":
            self.is_cif = True
            self.xyz = None
            self.molecules = None
            self.mol = None
            self.nframes = 1
            self.cif = CifParser(self.file)
            self.struct = Structure.from_file(self.file)
            # self.struct = self.cif.get_structures()[0]  #: NOT WORKING!
            logger.debug("structure created from a CIF file")
        elif self.file_ext == ".xyz":
            self.is_xyz = True
            self.cif = None
            self.xyz = XYZ.from_file(self.file)
            self.molecules = self.xyz.all_molecules
            self.mol = self.molecules[self.frame]
            self.nframes = len(self.molecules)
            self.struct = xyz2struct(self.mol)
            logger.debug("structure created from a XYZ file")
        elif self.file_ext == ".mpjson":
            self.is_xyz = False
            self.is_cif = True
            self.molecules = None
            self.mol = None
            self.nframes = 1
            self.struct = Structure.from_dict(json.load(open(self.file, "r")))
            logger.debug("structure created from JSON file")

        else:
            errmsg = "only CIF and XYZ files are currently supported"
            logger.error(errmsg)
            raise NotImplementedError(errmsg)

    def _get_atom_sites(self):
        """get atomic positions from the cif file

        Returns
        -------
        atom_sites : list of list
            same output as self.get_abs_sites()
        """
        if not self.is_cif:
            errmsg = "not a CIF file!"
            logger.error(errmsg)
            raise NotImplementedError(errmsg)
        cf = self.cif.as_dict()
        cf = cf[list(cf.keys())[0]]

        atom_lists = []
        try:
            labels = cf["_atom_site_label"]
            natoms = len(labels)
            try:
                type_symbols = cf["_atom_site_type_symbol"]
            except KeyError:
                type_symbols = None
                pass
            if type_symbols is None:
                atom_lists.append(labels)
            else:
                atom_lists.append(type_symbols)
            atom_lists.append(cf["_atom_site_fract_x"])
            atom_lists.append(cf["_atom_site_fract_y"])
            atom_lists.append(cf["_atom_site_fract_z"])
            atom_lists.append(cf["_atom_site_occupancy"])
        except KeyError:
            errmsg = (
                "the CIF file does not contain all required `_atom_site_*` information"
            )
            logger.error(errmsg)
            raise KeyError(errmsg)
        try:
            atom_lists.append(cf["_atom_site_symmetry_multiplicity"])
        except KeyError:
            atom_lists.append(["?"] * natoms)

        try:
            atom_lists.append(cf["_atom_site_Wyckoff_symbol"])
        except KeyError:
            atom_lists.append(["?"] * natoms)

        atom_sites = []
        for ikey, key in enumerate(zip(*atom_lists)):
            atom_row = [
                ikey,
                key[0],
                np.array(
                    [key[1].split("(")[0], key[2].split("(")[0], key[3].split("(")[0]],
                    dtype=float,
                ),
                key[5] + key[6],
                np.array([None, None, None]),
                float(key[4]),
                None,
            ]
            atom_sites.append(atom_row)

        # atom_site_keys = [key for key in cf.keys() if '_atom_site' in key]
        # atom_lists = [cf[key] for key in atom_site_keys]
        # atom_sites = []
        # for ikey, key in enumerate(zip(*atom_lists)):
        #    atom_row = [ikey, key[1], np.array([key[4], key[5], key[6]], dtype=float), key[2] + key[3], np.array([None, None, None]), float(key[8]), None]
        #    atom_sites.append(atom_row)
        return atom_sites

    def _get_idx_struct(self, atom_coords):
        """get the index of the pymatgen Structure corresponding to the given atomic coordinates"""
        for idx, atom in enumerate(self.struct):
            if np.allclose(atom.coords, atom_coords, atol=0.001) is True:
                return idx
        errmsg = f"atomic coordinates {atom_coords} not found in self.struct"
        logger.error(errmsg)
        # raise IndexError(errmsg)
        return None

    def get_abs_sites(self):
        """
        Get information about the possible absorbing sites present of the
        structure.

        ..note:: If the structure has a readable symmetry given by a cif file,
        the method will return just equivalent sites. If the structure does not
        have symmetry or the symmetry is not explicit in the files, this method
        will return all possible sites for absorber atoms.

        Returns
        -------
        abs_sites : list of lists
            The lists inside the list contain the following respective
            information:
            [
            abs_idx,        # index identifier of the absorber site
                             #     used by self.set_abs_site().
            species_string,  #  The specie for absorber sites.
            frac_coords,     # Fractional coordinate position
                             #   If the structure was created using xyz file,
                             #   the frac. coords. are arbitrary and the lattice
                             #   parameters are based on the molecule size.
            wyckoff_symbols, # Wyckoff site for absorber sites.
                             #      For structures created from xyz
                             #      files, Wyckoff sites are always equal to 1a.
                             #      (No symmetry)
            cart_coords,     # Cartesian coordinate position
            occupancy,       # Occupancy for absorber sites. For structures
                             #      created from xyz files,
                             #      occupancy are always equal to 1.
            structure index  # Original index in the pymatgen structure
                             #  (private usage)
            ]
        """

        abs_sites = []
        if self.is_cif:
            sym_struct = SpacegroupAnalyzer(self.struct).get_symmetrized_structure()

            # Get multiples sites for absorber atom
            for idx, sites in enumerate(sym_struct.equivalent_sites):
                sites = sorted(
                    sites, key=lambda s: tuple(abs(x) for x in s.frac_coords)
                )
                site = sites[0]
                abs_row = [idx, site.species_string]
                abs_row.append([j for j in np.round(site.frac_coords, 4)])
                abs_row.append(sym_struct.wyckoff_symbols[idx])
                abs_row.append(np.array([j for j in np.round(site.coords, 4)]))
                if self.abs_atom in abs_row[1]:
                    try:
                        ats_occ = abs_row[1].split(",")
                        at_occ = [at for at in ats_occ if self.abs_atom in at][0]
                        occupancy = float(at_occ.split(":")[1])
                        self.full_occupancy = False
                    except Exception:
                        occupancy = 1
                        self.full_occupancy = True
                    abs_row.append(occupancy)
                    abs_row.append(self._get_idx_struct(abs_row[4]))
                    abs_sites.append(abs_row)

        if self.is_xyz:
            self.full_occupancy = True
            k = 0
            for idx, elem in enumerate(self.struct):
                if f"{self.abs_atom}" in elem:
                    abs_sites += [
                        (
                            int(f"{k}"),
                            (str(elem.specie)) + f"{k}",
                            elem.coords.round(4),
                            "1a",
                            elem.coords.round(4),
                            1,
                            idx,
                        )
                    ]
                    k += 1
        if len(abs_sites) == 0:
            _errmsg = f"---- Absorber {self.abs_atom} not found ----"
            logger.error(_errmsg)
            raise AttributeError(_errmsg)
        return abs_sites

    def get_abs_sites_info(self):
        """pretty print for self.get_abs_sites()"""
        header = [
            "idx_abs",
            "specie",
            "frac_coords",
            "wyckoff_site",
            "cart_coords",
            "occupancy",
            "idx_in_struct",
        ]
        abs_sites = self.get_abs_sites()
        if HAS_PANDAS:
            df = pd.DataFrame(
                abs_sites,
                columns=header,
            )
            df = Styler(df).hide(axis="index")
            return df
        else:
            matrix = [header]
            matrix.extend(abs_sites)
            _pprint(matrix)

    def get_atoms_from_abs(self, radius):
        """Get atoms in sphere from absorbing atom with certain radius"""
        abs_sites = self.get_abs_sites()

        if self.is_cif:
            nei_list = self.struct.get_sites_in_sphere(
                abs_sites[self.abs_site][4], float(radius)
            )  # list os neighbors atoms in sphere
            sites = []

            for i in range(len(nei_list)):
                nei_list[i].cart_coords = (
                    nei_list[i].coords - abs_sites[self.abs_site][4]
                )

            for i, _ in enumerate(nei_list):
                if np.allclose(nei_list[i].cart_coords, [0, 0, 0], atol=0.01) is True:
                    sites.append(
                        [
                            nei_list[i],
                            f"{self.abs_atom}(abs)",
                            0.000,
                            {f"{self.abs_atom}(abs)": 1},
                        ]
                    )
                    nei_list.remove(nei_list[i])

            for i in range(len(nei_list)):
                occu_dict = dict(nei_list[i].as_dict()["species"])
                sites += [
                    [
                        nei_list[i],
                        str(nei_list[i].species.elements[0]),
                        round(np.linalg.norm(nei_list[i].cart_coords - [0, 0, 0]), 5),
                        occu_dict,
                    ]
                ]
        if self.is_xyz:
            nei_list = self.mol.get_sites_in_sphere(
                abs_sites[self.abs_site][4], float(radius)
            )  # list os neighbors atoms in sphere
            sites = []

            for i in range(len(nei_list)):
                nei_list[i].cart_coords = (
                    nei_list[i].coords - abs_sites[self.abs_site][4]
                )

            for i, _ in enumerate(nei_list):
                if np.allclose(nei_list[i].cart_coords, [0, 0, 0], atol=0.01) is True:
                    sites.append([nei_list[i], f"{self.abs_atom}(abs)", 0.000])
                    nei_list.remove(nei_list[i])

            sites += [
                [
                    nei_list[i],
                    nei_list[i].species_string,
                    round(np.linalg.norm(nei_list[i].cart_coords - [0, 0, 0]), 5),
                ]
                for i in range(len(nei_list))
            ]

        sites = sorted(sites, key=lambda x: x[2])
        return sites

    def get_coord_envs(self):
        """
        For structures from cif files, this method will try to find the
        coordination environment type and return the elements and the
        coordination env. symbol from the first using the classes from pymatgen
        as LocalGeometryFinder(), BVAnalyzer(), MultiWeightsChemenvStrategy()
        and LightStructureEnvironments() .

            > coordination env. symbol.
                        - `S:4` - Square Plane
                        - `T:4` - Tetrahedral
                        - `T:5` - Trigonal bipyramid
                        - `S:5` - Square pyramidal
                        - `O:6` - Octahedral
                        - `T:6` - Trigonal prism
            > ce_fraction:
                        probability for given coordination env. (between 0 and
                        1)

            > CSM:
                        a measure of the degree of symmetry in the coordination
                        environment. It is based on the idea that symmetric
                        environments are more stable than asymmetric ones, and
                        is calculated using a formula that takes into account
                        the distances and angles between the coordinating
                        atoms. The CSM can be understood as a distance to a
                        shape and can take values between 0.0 (if a given
                        environment is perfect) and 100.0 (if a given
                        environment is very distorted). The environment of the
                        atom is then the model polyhedron for which the
                        similarity is the highest, that is, for which the CSM
                        is the lowest.

            > permutation:
                        possible permutation of atoms surrounding the central
                        atom. This is a list that indicates the order in which
                        the neighboring atoms are arranged around the central
                        atom in the coordination environment. The numbering
                        starts from 0, and the list indicates the indices of
                        the neighboring atoms in this order. For example, in
                        the second entry of the list above, the permutation [0,
                        2, 3, 1, 4] means that the first neighboring atom is in
                        position 0, the second is in position 2, the third is
                        in position 3, the fourth is in position 1, and the
                        fifth is in position 4. The permutation is used to
                        calculate the csm value.

            > site:
                        element in the coordination environment and its
                        coordinates (cartesian and fractional).

            > site_index:
                        structure index for the coordinated atom.


        For structures from the xyz file the methods will try to return the
        elements (but not the coord. env. symbol) for the first coordination
        env. shell using the the class CrystalNN from pymatgen, which gives
        better results than BrunnerNN_real and CutOffDictNN (as previously
        tested)

        List of lists:

            [0]: Info about which site is being analyzed.

            [1]: Coord. env as dictionary.

            [2]: Info about coord. env.
                > site:
                    element in the coordination environment and its coordinates
                    (cartesian and fractional).

                > image:
                    image is defined as displacement from original site in
                    structure to a given site. i.e. if structure has a site at
                    (-0.1, 1.0, 0.3), then (0.9, 0, 2.3) -> jimage = (1, -1,
                    2). Note that this method takes O(number of sites) due to
                    searching an original site.

                > weight:
                    quantifies the significance or contribution of each
                    coordinated site to the central site's coordination.

                > site_index:
                    structure index for the coerdinated atom.


        """
        abs_sites = self.get_abs_sites()
        idx_abs_site = abs_sites[self.abs_site][-1]

        if self.is_cif:
            from pymatgen.analysis.bond_valence import BVAnalyzer
            from pymatgen.analysis.chemenv.coordination_environments.coordination_geometry_finder import (
                LocalGeometryFinder,
            )
            from pymatgen.analysis.chemenv.coordination_environments.structure_environments import (
                LightStructureEnvironments,
            )
            from pymatgen.analysis.chemenv.coordination_environments.chemenv_strategies import (
                # SimplestChemenvStrategy,
                MultiWeightsChemenvStrategy,
                # WeightedNbSetChemenvStrategy,
            )

            lgf = LocalGeometryFinder()
            lgf.setup_structure(self.struct)

            bva = BVAnalyzer()  # Bond Valence Analyzer
            try:
                valences = bva.get_valences(structure=self.struct)
            except ValueError:
                valences = "undefined"

            coord_env_list = []
            se = lgf.compute_structure_environments(
                max_cn=6,
                valences=valences,
                only_indices=[idx_abs_site],
                only_symbols=["S:4", "T:4", "T:5", "S:5", "O:6", "T:6"],
            )

            dist_1st_shell = se.voronoi.neighbors_distances[idx_abs_site][0]["max"]
            logger.debug(dist_1st_shell)
            # atom_coord = lgf.compute_coordination_environments(self.struct)
            strategy = MultiWeightsChemenvStrategy.stats_article_weights_parameters()
            # strategy = SimplestChemenvStrategy(distance_cutoff=1.1, angle_cutoff=0.3)
            lse = LightStructureEnvironments.from_structure_environments(
                strategy=strategy, structure_environments=se
            )
            coord_env_ce = lse.coordination_environments[idx_abs_site]
            ngbs_sites = lse._all_nbs_sites
            coord_env_list.append(
                [
                    f"Coord. Env. for Site {abs_sites[self.abs_site][0]}",
                    coord_env_ce,
                    ngbs_sites,
                ]
            )

        if self.is_xyz:
            from pymatgen.analysis.local_env import CrystalNN

            obj = CrystalNN()
            coord_env_list = []
            coord_env = obj.get_nn_info(self.struct, idx_abs_site)
            for site in coord_env:
                site["site"].cart_coords = self.struct[site["site_index"]].coords
            coord_dict = obj.get_cn_dict(self.struct, idx_abs_site)
            coord_env_list.append(
                [
                    f"Coord. Env. for Site {abs_sites[self.abs_site][0]}",
                    {"ce_symbol": f"Elements Dict = {coord_dict}"},
                    coord_env,
                ]
            )

        return coord_env_list

    def get_coord_envs_info(self):
        """
        Class with summarized and more readable information from get_coord_envs() method
        """

        coord_env = self.get_coord_envs()[0]
        abs_site_coord = self.get_abs_sites()[self.abs_site][4]

        elems_dist = []
        for site in coord_env[2]:
            if self.is_cif:
                coord_sym = [
                    coord_env[1][i]["ce_symbol"] for i in range(len(coord_env[1]))
                ]
                elems_dist.append(
                    (
                        site["site"].species,
                        round(np.linalg.norm(site["site"].coords - abs_site_coord), 5),
                    )
                )
            if self.is_xyz:
                coord_sym = coord_env[1]["ce_symbol"]
                elems_dist.append(
                    (
                        site["site"].species,
                        round(
                            np.linalg.norm(site["site"].cart_coords - abs_site_coord), 5
                        ),
                    )
                )
        elems_dist = sorted(elems_dist, key=lambda x: x[1])
        print(
            f"Coord. Env. from absorber atom: {self.abs_atom} at site {self.get_abs_site()}"
        )
        print(coord_sym)
        header = ["Element", "Distance"]
        if HAS_PANDAS:
            df = pd.DataFrame(data=elems_dist, columns=header)
            return df
        else:
            matrix = [header]
            matrix.extend(elems_dist)
            _pprint(matrix)

    def make_cluster(self, radius):
        """Create a cluster with absorber atom site at the center.

        Arguments
        ---------
        radius :float
            cluster radius [Angstrom]

        Returns
        -------
        atoms : list
            species and coords for the new cluster structure
        """

        selected_site = self.get_abs_sites()[self.abs_site]
        cluster = self.mol.get_sites_in_sphere(selected_site[-3], radius)

        # showing and storing cartesian coords and species
        atoms = []

        # abs_atom at the cluster center
        for i in range((len(cluster))):
            try:
                species = round(Element((cluster[i].specie).element).Z)
            except AttributeError:
                species = round(Element(cluster[i].specie).Z)

            # getting species, after atomic number() and rounding
            coords = (
                cluster[i].coords - selected_site[2]
            )  # cartesial coords and ""frac_coords"" are the same for molecule structure (a = b = c = 1)
            coords = np.around(coords, 5)
            dist = round(np.linalg.norm(coords - [0, 0, 0]), 5)
            atoms.append((species, coords, dist))
        atoms = sorted(atoms, key=lambda x: x[2])
        return atoms

    def make_input_fdmnes(
        self, radius=7, parent_path=None, template=None, green=True, **kwargs
    ):
        """
        Create a fdmnes input from a template.

        Arguments
        ---------
        radius : float, [7]
            radius for fdmnes calculation in Angstrom
        parent_path : str, [None]
            path to the parent directory where the input files are stored
            if None it will create a temporary directory
        template : str, [None]
            full path to the template file
        green : boolean [True]
            True: use `Green` (muffin-tin potentials, faster)
            False: use finite-difference method (slower)

        ..note:: SCF is always used
        ..note:: for further information about FDMNES keywords, search for "FDMNES users guide"

        Returns
        -------
        None -> writes FDMNES input to disk
        directory structure: {parent_path}/fdmnes/{self.file_name}/{self.abs_atom}/frame{self.frame}/site{self.abs_site}/

        """
        replacements = {}
        replacements.update(**kwargs)
        replacements["version"] = __version__

        if template is None:
            template = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "templates", "fdmnes.tmpl"
            )

        if parent_path is None:
            parent_path = self.folders["fdmnes"]

        self.outdir = os.path.join(
            parent_path,
            self.file_name,
            self.abs_atom,
            f"frame{self.frame}",
            f"site{self.abs_site}",
        )

        method = "green" if green else ""
        absorber = ""
        crystal = ""
        occupancy = ""
        comment = f"input structure: {self.file_name}{self.file_ext}\ncreation date: {_get_timestamp()}"

        if self.is_cif:
            try:
                selected_site = self.get_abs_sites()[self.abs_site]
            except IndexError:
                logger.error("IndexError: check if abs_atom is correct")

            if not selected_site[-2] == 1:
                logger.warning("the selected site does not have full occupancy!")

            # SpacegroupAnalyzer to get symmetric structure
            analyzer = SpacegroupAnalyzer(self.struct)
            structure = analyzer.get_refined_structure()

            symmetry_data = analyzer.get_symmetry_dataset()
            group_number = symmetry_data["number"]
            group_choice = symmetry_data["choice"]

            # FDMNES doesn't recognize 2 as a space group.
            if group_number == 2:
                group_number = "P-1"

            crystal = f"{crystal}"
            replacements["crystal"] = "crystal"

            group = f"{group_number}"
            if group_choice:
                group += f":{group_choice}"
            replacements["group"] = f"spgroup\n{group}"

            unique_sites = []
            for sites in analyzer.get_symmetrized_structure().equivalent_sites:
                sites = sorted(
                    sites, key=lambda s: tuple(abs(x) for x in s.frac_coords)
                )
                unique_sites.append((sites[0], len(sites)))
                sites = str()
            if self.full_occupancy:
                for i, (site, _) in enumerate(unique_sites):
                    try:
                        e = site.specie
                    except AttributeError:
                        e = Element(site.species_string.split(":")[0])
                    sites += "\n" + (
                        f"{e.Z:>2d} {site.a:12.8f} {site.b:12.8f} {site.c:12.8f}"
                        f" {site.species_string:>4s}"
                    )
            else:
                occupancy = "occupancy"
                for site, _ in unique_sites:
                    for i, e in enumerate(site.species.elements):
                        occu = site.as_dict()["species"][i]["occu"]
                        sites += "\n" + (
                            f"{e.Z:>2d} {site.a:12.8f} {site.b:12.8f} {site.c:12.8f} {occu}"
                            f" {str(e):>4s}"
                        )
            lat = structure.lattice
            replacements["lattice"] = (
                f"{lat.a:<12.8f} {lat.b:12.8f} {lat.c:12.8f} "
                f"{lat.alpha:12.8f} {lat.beta:12.8f} {lat.gamma:12.8f}"
            )

            absorber = f"{absorber}"
            for i in range(len(unique_sites)):
                if (
                    np.allclose(unique_sites[i][0].coords, selected_site[4], atol=0.01)
                    is True
                ):
                    replacements["absorber"] = f"absorber\n{i+1}"

            # absorber = f"{absorber}"
            # replacements["absorber"] = f"Z_absorber\n{round(Element(elem).Z)}"

        if self.is_xyz:
            replacements["crystal"] = "molecule"

            atoms = self.make_cluster(radius=radius)
            sites = str()
            for i in range(len(atoms)):
                e = atoms[i][0]
                c = atoms[i][1]
                sites += "\n" + (
                    f"{e:>2d} {c[0]:12.8f} {c[1]:12.8f} {c[2]:12.8f}"
                    f" {Element.from_Z(e).name}"
                )

            absorber = f"{absorber}"
            for i in range(len(atoms)):
                if np.allclose(atoms[i][1], [0, 0, 0], atol=0.01) is True:
                    replacements["absorber"] = f"absorber\n{i+1}"

            replacements["group"] = ""

            lat = self.struct.lattice
            replacements["lattice"] = (
                f"{lat.a:<12.8f} {lat.b:12.8f} {lat.c:12.8f} "
                f"{lat.alpha:12.8f} {lat.beta:12.8f} {lat.gamma:12.8f}"
            )

        replacements["sites"] = sites
        replacements["radius"] = radius
        replacements["method"] = method
        replacements["comment"] = comment
        replacements["occupancy"] = occupancy

        try:
            os.makedirs(self.outdir, mode=0o755)
        except FileExistsError:
            pass

        # Write the input file.
        fnout = os.path.join(self.outdir, "job_inp.txt")
        with open(fnout, "w") as fp, open(template) as tp:
            inp = tp.read().format(**replacements)
            fp.write(inp)

        # Write the fdmfile.txt.
        with open(os.path.join(self.outdir, "fdmfile.txt"), "w") as fp:
            fp.write("1\njob_inp.txt")

        logger.info(f"written FDMNES input -> {fnout}")

    def make_input_feff(
        self,
        radius=7,
        parent_path=None,
        template=None,
        feff_comment="*",
        edge="K",
        sig2=None,
        debye=None,
        **kwargs,
    ):
        """
        Create a FEFF input from a template.

        Arguments
        ---------
        radius : float, [7]
            radius for feff calculation [Angstrom].
        parent_path : str, [None]
            path to the parent directory where the input files are stored
            if None it will create a temporary directory
        template : str, [None]
            full path to the template file
        feff_coment : str, ["*"]
            comment character used in the input file
        sig2 : float or None, [None]
            SIG2 keywork, if None it will be commented
        debye : list of two floats or None, [None]
            DEBYE keyword, if None it will be commented, otherwise:
            debye=[temperature, debye_temperature]
                temperatue : float
                    temperature at which the Debye-Waller factors are calculated [Kelvin].
                debye_temperature : float
                    Debye Temperature of the material [Kelvin].

        ..note:: refer to [FEFF documentation](https://feff.phys.washington.edu/feffproject-feff-documentation.html)

        Returns
        -------
        None -> writes FEFF input to disk
        directory structure: {parent_path}/feff/{self.file_name}/{self.abs_atom}/frame{self.frame}/site{self.abs_site}/
        """
        replacements = {}
        replacements.update(**kwargs)
        replacements["version"] = __version__

        if parent_path is None:
            parent_path = self.folders["feff"]
        self.outdir = os.path.join(
            parent_path,
            self.file_name,
            self.abs_atom,
            f"frame{self.frame}",
            f"site{self.abs_site}",
        )

        if template is None:
            template = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "templates",
                "feff_exafs.tmpl",
            )

        if sig2 is None:
            use_sig2 = "*"
            sig2 = 0.005
        else:
            use_sig2 = ""

        if debye is None:
            use_debye = "*"
            temperature, debye_temperature = 0, 0
        else:
            use_debye = ""
            temperature, debye_temperature = debye[0], debye[1]

        feff_comment = f"{feff_comment}"
        edge = f"{edge}"
        radius = f"{radius}"
        use_sig2 = f"{use_sig2}"
        sig2 = f"{sig2}"
        use_debye = f"{use_debye}"
        temperature = f"{temperature}"
        debye_temperature = f"{debye_temperature}"

        if self.is_cif:
            sites = self.get_atoms_from_abs(radius)
            ipot_list = [(0, Element(f"{self.abs_atom}").Z, sites[0][1])]
            ipot = {f"{self.abs_atom}(abs)": 0}
            elems = self.species

            for i, elem in enumerate(elems):
                ipot[elem] = i + 1
            for i in range(1, len(sites)):
                for j, _ in enumerate(sites[i][0].species.elements):
                    if str(sites[i][0].species.elements[j]) in elems:
                        ipot_list.append(
                            (
                                ipot[str(sites[i][0].species.elements[j])],
                                sites[i][0].species.elements[j].Z,
                                str(sites[i][0].species.elements[j]),
                            )
                        )
            pot = list(dict.fromkeys(ipot_list))
            potentials = str("* ipot  Z   tag [lmax1 lmax2 xnatph sphinph]")
            for i, _ in enumerate(pot):
                potentials += "\n" + (f"{pot[i][0]:>5} {pot[i][1]:>3} {pot[i][2]:>5}")

            atoms_list = [
                (sites[0][0].cart_coords, 0, sites[0][1], sites[0][2], sites[0][3])
            ]
            for i in range(1, len(sites)):
                atoms_list.append(
                    (
                        sites[i][0].cart_coords,
                        ipot[str(sites[i][0].species.elements[0])],
                        sites[i][1],
                        sites[i][2],
                        sites[i][3],
                    )  # cart, ipot, tag, dist
                )

        if self.is_xyz:
            sites = self.get_atoms_from_abs(radius=radius)
            ipot_list = [(0, Element(f"{self.abs_atom}").Z, sites[0][1])]
            elems = []
            ipot = {}
            for i in range(1, len(sites)):
                if sites[i][0].species_string not in elems:
                    elems.append((sites[i][0].species_string))
            for i, elem in enumerate(elems):
                ipot[elem] = i + 1
            for i in range(1, len(sites)):
                if sites[i][0].species_string in elems:
                    ipot_list.append(
                        (
                            ipot[sites[i][0].species_string],
                            sites[i][0].specie.Z,
                            sites[i][0].species_string,
                        )
                    )
            pot = list(dict.fromkeys(ipot_list))
            potentials = str("* ipot  Z   tag [lmax1 lmax2 xnatph sphinph]")
            for i in range(len(pot)):
                potentials += "\n" + (f"{pot[i][0]:>5} {pot[i][1]:>3} {pot[i][2]:>5}")
            atoms_list = []
            for i in range(len(sites)):
                atoms_list.append(
                    (
                        sites[i][0].cart_coords,
                        ipot_list[i][0],
                        sites[i][1],  # tag
                        sites[i][2],
                    )
                )
        atoms = str(
            "*   x          y          z     ipot   tag    distance   occupancy"
        )
        at = atoms_list
        for i in range(len(at)):
            if self.full_occupancy:
                atoms += "\n" + (
                    f"{at[i][0][0]:10.6f} {at[i][0][1]:10.6f} {at[i][0][2]:10.6f} {  int(at[i][1])}  {at[i][2]:>5} {at[i][3]:10.5f}         *1 "
                )
            else:
                choice = np.random.choice(
                    list(at[i][4].keys()), p=list(at[i][4].values())
                )
                atoms += "\n" + (
                    f"{at[i][0][0]:10.6f} {at[i][0][1]:10.6f} {at[i][0][2]:10.6f} {ipot[choice]}  {choice:>5} {at[i][3]:10.5f} *{at[i][4]}"
                )

        title = f"TITLE {self.file_name}{self.file_ext}\nTITLE {_get_timestamp()}\nTITLE site {self.abs_site}"

        replacements["feff_comment"] = feff_comment
        replacements["edge"] = edge
        replacements["radius"] = radius
        replacements["use_sig2"] = use_sig2
        replacements["sig2"] = sig2
        replacements["use_debye"] = use_debye
        replacements["temperature"] = temperature
        replacements["debye_temperature"] = debye_temperature
        replacements["potentials"] = potentials
        replacements["atoms"] = atoms
        replacements["title"] = title
        # replacements[""] =

        try:
            os.makedirs(self.outdir, mode=0o755)
        except FileExistsError:
            pass

        # Write the input file.
        fnout = os.path.join(self.outdir, "feff.inp")
        with open(fnout, "w") as fp, open(template) as tp:
            inp = tp.read().format(**replacements)
            fp.write(inp)

        logger.info(f"written FEFF input -> {fnout}")

    def _get_xyz_and_elements(self, radius):
        """
        Get information about cartesian coords and elements surrounding the central atom given
        a radius.

         Args:
            > radius(float):
                    radius from the central atom [Angstrom].

         return list of elements with coords and list of elements, both lists of strings.

        """
        sites = self.get_atoms_from_abs(radius)
        coords = []
        elements = []
        for _, site in enumerate(sites):
            try:
                coords.append(
                    (str((site[0].species).elements[0].name), site[0].cart_coords)
                )
            except AttributeError:
                coords.append((str(site[0].specie), site[0].cart_coords))

        output_str = str(len(coords)) + "\n\n"
        for element, coords in coords:
            coords_str = " ".join([f"{c:.6f}" for c in coords])
            output_str += f"{element} {coords_str}\n"
            if element not in elements:
                elements.append(element)
        elements = sorted(elements)
        return output_str, elements

    def _round_up(self, x):
        rounded_x = np.ceil(x * 100) / 100
        return rounded_x

    def visualize(self, radius=2.5, unitcell=False):
        """
        Display a 3D visualization for material local structure.

        Args:
            > radius (float):
                    radius visualization from the central atom.

            > unitcell (boolean):
                    if True, allows the visualization of the structure unit cell.

        return 3D structure visualization from py3Dmol.
        """
        if HAS_PY3DMOL is False:
            logger.error("py3Dmol not installed! -> run `pip install py3Dmol`")
            return

        radius = self._round_up(radius)

        xyz, elems = self._get_xyz_and_elements(radius)

        a = self.struct.lattice.a
        b = self.struct.lattice.b
        c = self.struct.lattice.c
        alpha = self.struct.lattice.alpha
        beta = self.struct.lattice.beta
        gamma = self.struct.lattice.gamma
        xyzview = py3Dmol.view(
            width=600, height=600
        )  # http://3dmol.org/doc/GLViewer.html#setStyle
        xyzview.addModel(xyz, "xyz")

        if unitcell is True:
            m = xyzview.getModel()
            m.setCrystData(a, b, c, alpha, beta, gamma)
            xyzview.addUnitCell()

        colors = [
            "red",
            "green",
            "blue",
            "orange",
            "yellow",
            "white",
            "purple",
            "pink",
            "brown",
            "black",
            "gray",
            "cyan",
            "magenta",
            "olive",
            "navy",
            "teal",
            "maroon",
            "turquoise",
            "indigo",
            "salmon",
        ]
        color_elems = {}
        for idx, elem in enumerate(self.elems):
            color_elems[f"{elem}"] = colors[idx]

        for idx, elem in enumerate(elems):
            color = color_elems[f"{elem}"]
            xyzview.setStyle(
                {"elem": f"{elem}"},
                {
                    "stick": {
                        "radius": 0.1,
                        "opacity": 1,
                        "hidden": False,
                        "color": f"{color}",
                    },
                    "sphere": {"color": f"{color}", "radius": 0.4, "opacity": 1},
                },
            )
        xyzview.addLabel(
            "Abs",
            {
                "fontColor": "black",
                "fontSize": 14,
                "backgroundColor": "white",
                "backgroundOpacity": 0.8,
                "showBackground": True,
            },
            {"index": 0},
        )

        xyzview.zoomTo()
        xyzview.show()

        logger.info(color_elems)
        if not self.full_occupancy:
            logger.warning("3D displayed image does not consider partial occupancy")
            logger.info("check atoms occupancy here:", self.atoms_occu)
            # color_elems = {}
            # for idx, elem in enumerate(self.species_occu):
            #     color_elems[f"{list(elem.values())[0]}"] = colors[idx]
            # print("Label:\n", color_elems)


def get_fdmnes_info(file, labels=("energy", "mu")):
    """Get info from the fdmnes output file such as edge energy, atomic number Z,
      and fermi level energy, and returns a group with the storage information

      Parameters:

        file (str): path to the fdmnes output file.
    Obs: The INPUT file must have the "Header" keyword to use this function in the OUTPUT file

    """
    group = read_ascii(file, labels=labels)

    with open(group.path) as f:
        line = f.readlines()[3]
        header = line.split()
        (
            e_edge,
            Z,
            e_fermi,
        ) = (float(header[0]), float(header[1]), float(header[6]))
        print(
            f"Calculated Fermi level: {e_fermi}\nAtomic_number: {Z}\nEnergy_edge: {e_edge}"
        )

    group.e_edge = e_edge
    group.Z = Z
    group.e_fermi = e_fermi

    return group


def convolve_data(
    energy, mu, group, fwhm=1, linbroad=[1.5, 0, 50], kernel="gaussian", efermi=None
):
    """
    Function for manual convolution using Convolution1D from larch and returning a group

    Generic discrete convolution

    Description
    -----------

    This is a manual (not optimized!) implementation of discrete 1D
    convolution intended for spectroscopy analysis. The difference with
    commonly used methods is the possibility to adapt the convolution
    kernel for each convolution point, e.g. change the FWHM of the
    Gaussian kernel as a function of the energy scale.

    Resources
    ---------

    .. [WPconv] <http://en.wikipedia.org/wiki/Convolution#Discrete_convolution>
    .. [Fisher] <http://homepages.inf.ed.ac.uk/rbf/HIPR2/convolve.htm>
    .. [GP1202] <http://glowingpython.blogspot.fr/2012/02/convolution-with-numpy.html>

    """

    gamma_e = lin_gamma(energy, fwhm=fwhm, linbroad=linbroad)
    mu_conv = conv(energy, mu, kernel=kernel, fwhm_e=gamma_e, efermi=efermi)
    group.conv = mu_conv
    return group


def save_cif_from_mp(api_key, material_id, parent_path=None):
    """Collect a CIF file from the Materials Project Database, given the material id

    Parameters
    ----------
    api_key : str
        api-key from Materials Project
    material id : str
        material id (format mp-xxxx) from Materials Project
    parent_path : str
        path where to store the CIF files
        if None, a temporary one is created

    Returns
    -------
        [str, str] : parent_path, CIF file name

    """
    if parent_path is None:
        parent_path = structure_folders()["mp_structs"]

    from pymatgen.ext.matproj import _MPResterLegacy

    cif = _MPResterLegacy(api_key).get_data(material_id, prop="cif")
    pf = _MPResterLegacy(api_key).get_data(material_id, prop="pretty_formula")[0][
        "pretty_formula"
    ]
    outfn = f"{pf}_{material_id}.cif"
    with open(file=os.path.join(parent_path, outfn), mode="w") as f:
        f.write(cif[0]["cif"])
        logger.info(f"{material_id} -> {outfn}")
    return [parent_path, outfn]


def save_mp_structure(api_key, material_id, parent_path=None):
    """Save structure from Materials Project Database as json, given the material id

    Parameters
    ----------
    api_key : str
        api-key from Materials Project
    material id : str
        material id (format mp-xxxx) from Materials Project
    parent_path : str
        path where to store the Structure files
        if None, user_larchdir + 'mp_structs' is used

    Returns
    -------
        name of structure file, which will have an 'mpjson' extension

    Notes
    ------
    The structure is saved as json that can be loaded with
          from pymatgen.core import Structure
          import json
          struct = Structure.from_dict(json.load(open(filename, 'r')))

    """

    if parent_path is None:
        parent_path = structure_folders()["mp_structs"]

    try:
        from mp_api.client import MPRester
    except ImportError:
        print("need to install mp_api:  pip install mp_api")

    mpr = MPRester(api_key)
    results = mpr.summary.search(
        material_ids=[material_id], fields=["structure", "formula_pretty"]
    )
    formula = results[0].formula_pretty
    structure = results[0].structure

    outfile = os.path.join(parent_path, f"{formula}_{material_id}.mpjson")
    with open(outfile, "w") as fh:
        fh.write(structure.to_json())
    logger.info(f"saved {material_id} to {outfile}")

    return outfile
